# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolve the Patron referenced in the DocumentRequest."""

import jsonresolver
from werkzeug.routing import Rule

from invenio_app_ils.patrons.api import get_patron_or_unknown_dump
from invenio_app_ils.records.jsonresolvers.api import (
    get_field_value_for_record as get_field_value,
)

from ..api import DocumentRequest

# Note: there must be only one resolver per file,
# otherwise only the last one is registered


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the referred Patron for an DocumentRequest record."""
    from flask import current_app

    def patron_resolver(document_request_pid):
        """Get the Patron record for the given DocumentRequest or raise."""
        try:
            patron_pid = get_field_value(
                DocumentRequest, document_request_pid, "patron_pid"
            )
        except KeyError:
            return {}

        return get_patron_or_unknown_dump(patron_pid)

    url_map.add(
        Rule(
            "/api/resolver/document-requests/<document_request_pid>/patron",
            endpoint=patron_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""ILS Document search APIs."""

from invenio_search import RecordsSearch
from invenio_search.api import DefaultFilter

from invenio_app_ils.search_permissions import search_filter_record_permissions


class DocumentSearch(RecordsSearch):
    """RecordsSearch for documents."""

    boosted_fields = [
        "identifiers.value.text^12.0",
        "identifiers.value^12.0",
        "title^8.0",
        "authors.full_name^6.0",
        "imprint.publisher^4.0",
        "edition^4.0",
        "keywords^2.0",
        "abstract^2.0",
    ]

    class Meta:
        """Search only on documents index."""

        index = "documents"
        doc_types = None
        default_filter = DefaultFilter(search_filter_record_permissions)

    def search_by_pid(self, *pids):
        """Retrieve documents with the given pid(s)."""
        return self.filter("terms", pid=pids)



# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Admin panel for sent notifications."""

from flask import Blueprint
from flask_admin.contrib.sqla import ModelView
from invenio_i18n import gettext as _

from invenio_app_ils.notifications.models import NotificationsLogs


class NotificationsLogsModelView(ModelView):
    """Invenio admin view for sent notifications."""

    column_list = (
        "id",
        "is_manually_triggered",
        "recipient_user_id",
        "triggered_on",
        "send_log",
        "message_id",
        "action",
        "pid_type",
        "pid_value",
    )

    # Entries are read-only
    can_create = False
    can_edit = False
    can_delete = False

    can_view_details = True

    form_excluded_columns = NotificationsLogs.__table__.columns
    column_default_sort = (NotificationsLogs.triggered_on, True)

    column_filters = column_list


blueprint = Blueprint(
    "ils_admin",
    __name__,
    template_folder="templates",
    static_folder="static",
)

notifications_logs = {
    "model": NotificationsLogs,
    "modelview": NotificationsLogsModelView,
    "name": "Notifications Logs",
    "category": _("User Management"),
}

__all__ = ("notifications_logs",)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test ILL borrowing requests actions."""

import json
from datetime import timedelta

import arrow
from flask import url_for

from invenio_app_ils.ill.api import BORROWING_REQUEST_PID_TYPE, BorrowingRequest
from tests.helpers import user_login

_HTTP_OK = [200, 201, 204]
ITEM_ENDPOINT = "invenio_records_rest.illbid_item"
LIST_ENDPOINT = "invenio_records_rest.illbid_list"


def _create_loan_action(pid, data, client, json_headers):
    """Send request to create a new loan from the ILL."""
    url = url_for("invenio_app_ils_ill.illbid_create_loan", pid_value=pid)
    data["transaction_location_pid"] = "locid-1"
    return client.post(url, headers=json_headers, data=json.dumps(data))


def _assert_create_loan_action_fails(pid, data, client, json_headers):
    """Asserts that the create loan action validation fails."""
    res = _create_loan_action(pid, data, client, json_headers)
    assert res.status_code == 400


def test_brwreq_create_loan_only_backoffice(client, testdata, json_headers, users):
    """Test that patron have no permission of creating loans from ILLs."""
    user_login(client, "patron1", users)
    res = _create_loan_action("illbid-2", dict(), client, json_headers)
    assert res.status_code == 403


def test_brwreq_create_loan_fails_on_wrong_status(
    db, client, testdata, json_headers, users
):
    """Test borrowing requests create loan action fails on wrong status."""
    user_login(client, "librarian", users)

    def _create_new_brwreq(data=None):
        brwreq = data or dict(
            document_pid="docid-3",
            provider_pid="ill-provid-2",
            patron_pid="1",
            status="PENDING",
            type="PHYSICAL_COPY",
        )
        url = url_for(LIST_ENDPOINT)
        res = client.post(url, headers=json_headers, data=json.dumps(brwreq))
        assert res.status_code in _HTTP_OK
        brw_req = res.get_json()["metadata"]
        return brw_req["pid"]

    def _update_brwreq_with_new_status(pid, status):
        rec = BorrowingRequest.get_record_by_pid(pid)
        rec["status"] = status
        if status == "CANCELLED":
            rec["cancel_reason"] = "OTHER"
        rec.commit()
        db.session.commit()

    def _assert_fail_when_status(pid, status):
        _update_brwreq_with_new_status(pid, status)

        now = arrow.utcnow()
        future = now + timedelta(days=15)
        data = dict(loan_start_date=now.isoformat(), loan_end_date=future.isoformat())
        _assert_create_loan_action_fails(pid, data, client, json_headers)

    pid = _create_new_brwreq()

    _assert_fail_when_status(pid, "PENDING")
    _assert_fail_when_status(pid, "ON_LOAN")
    _assert_fail_when_status(pid, "RETURNED")
    _assert_fail_when_status(pid, "CANCELLED")


def test_brwreq_create_loan_fails_on_wrong_loan_end_date(
    db, client, testdata, json_headers, users
):
    """Test borrowing requests create loan action fails on wrong end date."""
    user_login(client, "librarian", users)

    # demo data "illbid-2" has the valid state `REQUESTED`
    pid = "illbid-2"
    # missing date param
    _assert_create_loan_action_fails(pid, dict(), client, json_headers)

    now = arrow.utcnow()

    # wrong format date
    future = now + timedelta(days=5)
    data = dict(
        loan_start_date=now.isoformat(),
        loan_end_date=future.date().isoformat(),
    )
    _assert_create_loan_action_fails(pid, data, client, json_headers)
    data = dict(
        loan_start_date=now.date().isoformat(),
        loan_end_date=future.isoformat(),
    )
    _assert_create_loan_action_fails(pid, data, client, json_headers)

    # past date
    past = now - timedelta(days=15)
    data = dict(
        loan_start_date=now.date().isoformat(),
        loan_end_date=past.date().isoformat(),
    )
    _assert_create_loan_action_fails(pid, data, client, json_headers)


def test_brwreq_create_loan_fails_on_loan_pid_already_attached(
    db, client, testdata, json_headers, users
):
    """Test borrowing requests create loan action fails on loan_pid already."""
    user_login(client, "librarian", users)

    # demo data "illbid-2" has the valid state `REQUESTED`
    pid = "illbid-2"

    rec = BorrowingRequest.get_record_by_pid(pid)
    rec.setdefault("patron_loan", {})
    rec["patron_loan"]["pid"] = "loanid-3"
    rec.commit()
    db.session.commit()

    # already with a loan pid for some reasons
    now = arrow.utcnow()
    future = now + timedelta(days=5)
    data = dict(
        loan_start_date=now.date().isoformat(),
        loan_end_date=future.date().isoformat(),
    )
    _assert_create_loan_action_fails(pid, data, client, json_headers)


def test_brwreq_create_loan_succeeds(db, client, testdata, json_headers, users):
    """Test borrowing requests create loan action succeeds."""
    user = user_login(client, "librarian", users)

    # demo data "illbid-2" has the valid state `REQUESTED`
    pid = "illbid-2"

    now = arrow.utcnow()
    start = (now + timedelta(days=3)).date().isoformat()
    future = (now + timedelta(days=5)).date().isoformat()
    data = dict(loan_start_date=start, loan_end_date=future)
    res = _create_loan_action(pid, data, client, json_headers)
    assert res.status_code in _HTTP_OK
    brw_req = res.get_json()["metadata"]

    assert brw_req["status"] == "ON_LOAN"
    assert "patron_loan" in brw_req
    patron_loan = brw_req["patron_loan"]
    assert "pid" in patron_loan
    assert "loan" in patron_loan
    loan = patron_loan["loan"]
    assert patron_loan["pid"] == loan["pid"]
    assert loan["end_date"] == future
    loan_pid = loan["pid"]

    # fetch the loan
    url = url_for("invenio_records_rest.loanid_item", pid_value=loan_pid)
    res = client.get(url, headers=json_headers)
    assert res.status_code in _HTTP_OK
    loan = res.get_json()["metadata"]

    # make sure the loan is created with the data from the ILL
    assert loan["item_pid"] == dict(type=BORROWING_REQUEST_PID_TYPE, value=pid)
    assert loan["start_date"] == start
    assert loan["end_date"] == future
    assert loan["document_pid"] == brw_req["document_pid"]
    assert loan["patron_pid"] == brw_req["patron_pid"]
    assert loan["transaction_user_pid"] == str(user.id)

    # update notes (a random field)
    brw_req["notes"] = "This is a note"
    url = url_for(ITEM_ENDPOINT, pid_value=pid)
    res = client.put(url, headers=json_headers, data=json.dumps(brw_req))
    assert res.status_code in _HTTP_OK
    updated_brw_req = res.get_json()["metadata"]

    # make sure `patron_loan` system field is preserved
    assert updated_brw_req["notes"] == "This is a note"
    assert "patron_loan" in updated_brw_req
    assert "pid" in patron_loan
    assert "loan" in patron_loan



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Patron indexer APIs."""

from datetime import datetime

from celery import shared_task
from flask import current_app
from invenio_accounts.models import User
from invenio_circulation.pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from invenio_circulation.proxies import current_circulation
from invenio_circulation.search.api import search_by_patron_pid
from invenio_db import db
from invenio_indexer.api import RecordIndexer

from invenio_app_ils.acquisition.api import ORDER_PID_TYPE
from invenio_app_ils.acquisition.proxies import current_ils_acq
from invenio_app_ils.document_requests.api import DOCUMENT_REQUEST_PID_TYPE
from invenio_app_ils.ill.api import BORROWING_REQUEST_PID_TYPE
from invenio_app_ils.ill.proxies import current_ils_ill
from invenio_app_ils.indexer import ReferencedRecordsIndexer
from invenio_app_ils.patrons.api import PATRON_PID_TYPE
from invenio_app_ils.proxies import current_app_ils


def get_loans(patron_pid):
    """Get referenced loans."""
    referenced = []
    loan_record_cls = current_circulation.loan_record_cls
    for hit in search_by_patron_pid(patron_pid=patron_pid).scan():
        loan = loan_record_cls.get_record_by_pid(hit["pid"])
        referenced.append(dict(pid_type=CIRCULATION_LOAN_PID_TYPE, record=loan))
    return referenced


def get_document_requests(patron_pid):
    """Get referenced documents requests."""
    referenced = []
    docreq_search_cls = current_app_ils.document_request_search_cls
    docreq_record_cls = current_app_ils.document_request_record_cls
    for request in (
        docreq_search_cls().search_by_patron_pid(patron_pid=patron_pid).scan()
    ):
        docreq = docreq_record_cls.get_record_by_pid(request["pid"])
        referenced.append(dict(pid_type=DOCUMENT_REQUEST_PID_TYPE, record=docreq))
    return referenced


def get_acquisition_orders(patron_pid):
    """Get referenced acquisition orders."""
    referenced = []
    order_record_cls = current_ils_acq.order_record_cls
    order_search_cls = current_ils_acq.order_search_cls
    search = order_search_cls().search_by_patron_pid(patron_pid)
    for hit in search.scan():
        order = order_record_cls.get_record_by_pid(hit["pid"])
        referenced.append(dict(pid_type=ORDER_PID_TYPE, record=order))
    return referenced


def get_ill_borrowing_requests(patron_pid):
    """Get referenced ILL borrowing requests."""
    referenced = []
    brw_req_record_cls = current_ils_ill.borrowing_request_record_cls
    brw_req_search_cls = current_ils_ill.borrowing_request_search_cls
    search = brw_req_search_cls().search_by_patron_pid(patron_pid)
    for hit in search.scan():
        brw_req = brw_req_record_cls.get_record_by_pid(hit["pid"])
        referenced.append(dict(pid_type=BORROWING_REQUEST_PID_TYPE, record=brw_req))
    return referenced


@shared_task(ignore_result=True)
def index_referenced_records(patron):
    """Index referenced records."""
    indexer = ReferencedRecordsIndexer()

    patron_pid = patron["pid"]
    indexed = dict(pid_type=PATRON_PID_TYPE, record=patron)

    indexer.index(indexed, get_loans(patron_pid))
    indexer.index(indexed, get_document_requests(patron_pid))
    indexer.index(indexed, get_acquisition_orders(patron_pid))
    indexer.index(indexed, get_ill_borrowing_requests(patron_pid))


class PatronBaseIndexer(RecordIndexer):
    """Base indexer class for `Patron` to treat it as a record."""

    @staticmethod
    def _prepare_record(record, index, doc_type, arguments=None, **kwargs):
        """Prepare record data for indexing.

        :param record: The record to prepare.
        :param index: The Elasticsearch index.
        :param doc_type: The Elasticsearch document type.
        :returns: The record metadata.
        """
        data = record.dumps()
        return data

    def record_to_index(self, record):
        """Get index/doc_type given a record.

        :param record: The record where to look for the information.
        :returns: A tuple (index, doc_type).
        """
        doc_type = None
        return record._index, doc_type

    def index_by_id(self, record_uuid, **kwargs):
        """Not implemented."""
        raise NotImplementedError("Cannot use this operation for Patron")

    def delete_by_id(self, record_uuid, **kwargs):
        """Not implemented."""
        raise NotImplementedError("Cannot use this operation for Patron")

    def bulk_index(self, record_id_iterator):
        """Not implemented."""
        raise NotImplementedError("Cannot use this operation for Patron")

    def bulk_delete(self, record_id_iterator):
        """Not implemented."""
        raise NotImplementedError("Cannot use this operation for Patron")

    def process_bulk_queue(self, es_bulk_kwargs=None):
        """Not implemented."""
        raise NotImplementedError("Cannot use this operation for Patron")


class PatronIndexer(PatronBaseIndexer):
    """Indexer class for `Patron`."""

    def index(self, patron, arguments=None, **kwargs):
        """Index a Patron."""
        super().index(patron)
        eta = datetime.utcnow() + current_app.config["ILS_INDEXER_TASK_DELAY"]
        index_referenced_records.apply_async((patron,), eta=eta)

    def reindex_patrons(self):
        """Re-index all patrons."""
        # do not use PatronIndexer class otherwise it will trigger potentially
        # thousands of tasks to index referenced records
        indexer = PatronBaseIndexer()
        Patron = current_app_ils.patron_cls
        # cannot use bulk operation because Patron is not a real record
        all_user_ids = db.session.query(User.id).all()
        for (user_id,) in all_user_ids:
            patron = Patron(user_id)
            indexer.index(patron)
        return len(all_user_ids)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test errors."""

import pytest

from invenio_app_ils.errors import (
    ItemHasActiveLoanError,
    NotImplementedConfigurationError,
    PatronHasLoanOnItemError,
    PatronNotFoundError,
    RecordHasReferencesError,
    SearchQueryError,
    UnauthorizedSearchError,
)


def test_unauthorized_search_with_patron_pid(app):
    """Test UnauthorizedSearchError with patron_pid defined."""
    query = "test query"
    pid = "1"
    msg = "Search '{query}' not allowed by 'patron_pid:{patron_pid}'"
    with pytest.raises(UnauthorizedSearchError) as ex:
        raise UnauthorizedSearchError(query=query, patron_pid=pid)
    assert ex.value.code == 403
    assert ex.value.description == msg.format(query=query, patron_pid=pid)


def test_unauthorized_search_without_patron_pid(app):
    """Test UnauthorizedSearchError without patron_pid defined."""
    query = "test query"
    msg = "Search '{query}' not allowed by 'patron_pid:None'"
    with pytest.raises(UnauthorizedSearchError) as ex:
        raise UnauthorizedSearchError(query=query)
    assert ex.value.code == 401
    assert ex.value.description == msg.format(query=query)


def test_invalid_search_query(app):
    """Test SearchQueryError."""
    query = "invalid query"
    msg = "Invalid query syntax: '{query}'"
    with pytest.raises(SearchQueryError) as ex:
        raise SearchQueryError(query=query)
    assert ex.value.code == SearchQueryError.code
    assert ex.value.description == msg.format(query=query)


def test_not_implemented(app):
    """Test NotImplementedConfigurationError."""
    config_variable = "CONFIG_VAR"
    msg = (
        "Function is not implemented. Implement this function in your module "
        "and pass it to the config variable '{}'".format(config_variable)
    )
    with pytest.raises(NotImplementedConfigurationError) as ex:
        raise NotImplementedConfigurationError(config_variable="CONFIG_VAR")
    assert ex.value.code == NotImplementedConfigurationError.code
    assert ex.value.description == msg


def test_patron_not_found(app):
    """Test PatronNotFoundError."""
    patron_pid = "1"
    msg = "Patron with ID '{patron_pid}' was not found."
    with pytest.raises(PatronNotFoundError) as ex:
        raise PatronNotFoundError(patron_pid)
    assert ex.value.code == PatronNotFoundError.code
    assert ex.value.description == msg.format(patron_pid=patron_pid)


def test_patron_has_loan_on_item(app):
    """Test PatronHasLoanOnItemError."""
    patron_pid = "1"
    item_pid = dict(type="pitmid", value="2")
    msg = "Patron '{0}' has already an active loan on item '{1}:{2}'"
    with pytest.raises(PatronHasLoanOnItemError) as ex:
        raise PatronHasLoanOnItemError(patron_pid, item_pid)
    assert ex.value.code == PatronHasLoanOnItemError.code
    assert ex.value.description == msg.format(
        patron_pid, item_pid["type"], item_pid["value"]
    )


def test_record_has_references_error(app):
    """Test RecordHasReferencesError."""
    record_type = "Item"
    record_id = "item-1"
    ref_type = "RecordType"
    ref_ids = ["1", "2", "3"]
    msg = (
        "Cannot delete the {record_type} {record_id} record because it has "
        "references from {ref_type} records with ids: {ref_ids}."
    ).format(
        record_type=record_type,
        record_id=record_id,
        ref_type=ref_type,
        ref_ids=ref_ids,
    )

    with pytest.raises(RecordHasReferencesError) as ex:
        raise RecordHasReferencesError(
            record_type=record_type,
            record_id=record_id,
            ref_type=ref_type,
            ref_ids=ref_ids,
        )
    assert ex.value.code == RecordHasReferencesError.code
    assert ex.value.description == msg


def test_item_has_active_loan_error(app):
    loan_pid = "1"
    msg = (
        "Could not update item because it has an active loan with " "pid: {loan_pid}."
    ).format(loan_pid=loan_pid)
    with pytest.raises(ItemHasActiveLoanError) as ex:
        raise ItemHasActiveLoanError(loan_pid=loan_pid)
    assert ex.value.code == ItemHasActiveLoanError.code
    assert ex.value.description == msg



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test notifications backend emails."""

from invenio_app_ils.notifications.api import send_notification
from invenio_app_ils.notifications.messages import NotificationMsg
from invenio_app_ils.patrons.api import Patron


class FakeMessage(NotificationMsg):
    def __init__(self, *args, **kwargs):
        template = kwargs.pop("template", "notifications/title_body_html.html")
        super().__init__(template, **kwargs)


def test_send_only_to_test_recipients(app_with_notifs, users, testdata, mocker):
    """Tests that send only to test recipients works."""
    app_with_notifs.config["ILS_MAIL_ENABLE_TEST_RECIPIENTS"] = True
    succ = mocker.patch("invenio_app_ils.notifications.api.log_successful_notification")

    msg = FakeMessage()
    patron = Patron(users["patron1"].id)

    fake_recipients = app_with_notifs.config["ILS_MAIL_NOTIFY_TEST_RECIPIENTS"]
    with app_with_notifs.extensions["mail"].record_messages() as outbox:
        assert len(outbox) == 0
        send_notification([patron], msg)
        assert succ.s.called
        assert len(outbox) == 1
        assert outbox[0].recipients == fake_recipients

    app_with_notifs.config["ILS_MAIL_ENABLE_TEST_RECIPIENTS"] = False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Provider API."""

from functools import partial

from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2

from invenio_app_ils.acquisition.proxies import current_ils_acq
from invenio_app_ils.errors import RecordHasReferencesError
from invenio_app_ils.fetchers import pid_fetcher
from invenio_app_ils.ill.proxies import current_ils_ill
from invenio_app_ils.minters import pid_minter
from invenio_app_ils.records.api import IlsRecord

PROVIDER_PID_TYPE = "provid"
PROVIDER_PID_MINTER = "provid"
PROVIDER_PID_FETCHER = "provid"

ProviderIdProvider = type(
    "ProviderIdProvider",
    (RecordIdProviderV2,),
    dict(pid_type=PROVIDER_PID_TYPE, default_status=PIDStatus.REGISTERED),
)
provider_pid_minter = partial(pid_minter, provider_cls=ProviderIdProvider)
provider_pid_fetcher = partial(pid_fetcher, provider_cls=ProviderIdProvider)


class Provider(IlsRecord):
    """Provider class."""

    _pid_type = PROVIDER_PID_TYPE
    _schema = "providers/provider-v1.0.0.json"

    def delete(self, **kwargs):
        """Delete record."""
        order_search_cls = current_ils_acq.order_search_cls
        search_order_res = order_search_cls().search_by_provider_pid(self["pid"])

        brwreq_search_cls = current_ils_ill.borrowing_request_search_cls
        search_brwreq_res = brwreq_search_cls().search_by_provider_pid(self["pid"])

        if search_order_res.count() or search_brwreq_res.count():
            order_ref_ids = [res["pid"] for res in search_order_res.scan()]
            brwreq_ref_ids = [res["pid"] for res in search_brwreq_res.scan()]

            raise RecordHasReferencesError(
                record_type="Provider",
                record_id=self["pid"],
                ref_type="Order/BorrowingRequest",
                ref_ids=sorted(order_ref_ids + brwreq_ref_ids),
            )
        return super().delete(**kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test transitions on expiring loans."""

from datetime import timedelta

import arrow
from flask import url_for
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search

from invenio_app_ils.circulation.tasks import cancel_expired_loan_requests
from invenio_app_ils.patrons.api import SystemAgent
from tests.helpers import user_login


def test_cancel_expired_loans(client, json_headers, db, users, testdata):
    """Test cancellation of loan requests after expiration date."""

    user_login(client, "admin", users)

    def prepare_data():
        """Prepare data."""
        loans = testdata["loans"]

        recs = []
        now = arrow.utcnow()

        def new_expiration_date(loan, date):
            loan["request_expire_date"] = date.date().isoformat()
            loan["state"] = "PENDING"
            loan.commit()
            recs.append(loan)

        # expired loans
        date = now - timedelta(days=1)
        new_expiration_date(loans[0], date)
        new_expiration_date(loans[1], date)
        date = now - timedelta(days=2)
        new_expiration_date(loans[2], date)
        expired_pids = [loans[0]["pid"], loans[1]["pid"], loans[2]["pid"]]

        # not expired loans
        not_expired_pids = []
        remaining_not_expired = loans[3:]
        n_days = 0  # today
        for loan in remaining_not_expired:
            date = now + timedelta(days=n_days)
            new_expiration_date(loan, date)
            not_expired_pids.append(loan["pid"])
            n_days += 1
        db.session.commit()

        indexer = RecordIndexer()
        for rec in recs:
            indexer.index(rec)

        current_search.flush_and_refresh(index="*")
        return expired_pids, not_expired_pids

    expired_pids, not_expired_pids = prepare_data()

    # trigger cancel expired loan requests
    cancel_expired_loan_requests.apply_async()

    for pid_value in expired_pids:
        res = client.get(
            url_for("invenio_records_rest.loanid_item", pid_value=pid_value),
            headers=json_headers,
        )
        loan = res.get_json()["metadata"]
        assert loan["state"] == "CANCELLED"
        assert "cancel_reason" in loan
        assert loan["transaction_user_pid"] == str(SystemAgent.id)

    for pid_value in not_expired_pids:
        res = client.get(
            url_for("invenio_records_rest.loanid_item", pid_value=pid_value),
            headers=json_headers,
        )
        loan = res.get_json()["metadata"]
        assert loan["state"] == "PENDING"
        assert "cancel_reason" not in loan



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS Circulation APIs."""

import uuid
from copy import copy, deepcopy
from datetime import date, timedelta
from functools import partial

from flask import current_app
from flask_login import current_user
from invenio_circulation.config import (
    CIRCULATION_STATES_LOAN_ACTIVE,
    CIRCULATION_STATES_LOAN_COMPLETED,
)
from invenio_circulation.errors import CirculationException
from invenio_circulation.pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from invenio_circulation.proxies import current_circulation
from invenio_circulation.search.api import search_by_patron_item_or_document
from invenio_db import db
from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2

from invenio_app_ils.circulation.search import (
    get_all_expiring_or_overdue_loans_by_patron_pid,
)
from invenio_app_ils.errors import (
    DocumentOverbookedError,
    IlsException,
    InvalidLoanExtendError,
    InvalidParameterError,
    ItemCannotCirculateError,
    ItemHasActiveLoanError,
    ItemNotFoundError,
    MissingRequiredParameterError,
    MultipleItemsBarcodeFoundError,
    PatronHasLoanOnDocumentError,
    PatronHasLoanOnItemError,
    PatronHasRequestOnDocumentError,
)
from invenio_app_ils.fetchers import pid_fetcher
from invenio_app_ils.items.api import Item
from invenio_app_ils.minters import pid_minter
from invenio_app_ils.proxies import current_app_ils

# override default `invenio-circulation` minters to use the base32 PIDs
# CIRCULATION_LOAN_PID_TYPE is already defined in `invenio-circulation`
ILS_CIRCULATION_LOAN_MINTER = "ilsloanid"
ILS_CIRCULATION_LOAN_FETCHER = "ilsloanid"

IlsCirculationLoanIdProvider = type(
    "IlsCirculationLoanIdProvider",
    (RecordIdProviderV2,),
    dict(pid_type=CIRCULATION_LOAN_PID_TYPE, default_status=PIDStatus.REGISTERED),
)
ils_circulation_loan_pid_minter = partial(
    pid_minter, provider_cls=IlsCirculationLoanIdProvider
)
ils_circulation_loan_pid_fetcher = partial(
    pid_fetcher, provider_cls=IlsCirculationLoanIdProvider
)


def _validate_delivery(delivery):
    """Validate `delivery` param."""
    methods = list(
        current_app.config.get("ILS_CIRCULATION_DELIVERY_METHODS", {}).keys()
    )
    if methods:
        if not delivery or delivery["method"] not in methods:
            raise MissingRequiredParameterError(
                description="A valid 'delivery' is required on loan request"
            )


def _set_item_to_can_circulate(item_pid):
    """Change the item status to CAN_CIRCULATE."""
    item = Item.get_record_by_pid(item_pid["value"])
    if item["status"] != "CAN_CIRCULATE":
        item["status"] = "CAN_CIRCULATE"
        item.commit()
        db.session.commit()
        current_app_ils.item_indexer.index(item)


def patron_has_request_on_document(patron_pid, document_pid):
    """Return loan request for given patron and document."""
    states = current_app.config["CIRCULATION_STATES_LOAN_REQUEST"]
    search = search_by_patron_item_or_document(
        patron_pid=patron_pid, document_pid=document_pid, filter_states=states
    )
    search_result = search.execute()
    total = search_result.hits.total.value
    if total > 0:
        return search_result.hits[0]
    else:
        return None


def patron_has_active_loan_or_request_on_document(patron_pid, document_pid):
    """Return loan/request if it's active for the given patron and document."""
    states = (
        current_app.config["CIRCULATION_STATES_LOAN_REQUEST"]
        + current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
    )
    search = search_by_patron_item_or_document(
        patron_pid=patron_pid, document_pid=document_pid, filter_states=states
    )
    search_result = search.execute()
    total = search_result.hits.total.value
    if total > 0:
        return search_result.hits[0]
    else:
        return None


def request_loan(
    document_pid,
    patron_pid,
    transaction_location_pid,
    transaction_user_pid=None,
    **kwargs,
):
    """Create a new loan and trigger the first transition to PENDING."""
    loan_cls = current_circulation.loan_record_cls
    loan_or_request_found = patron_has_active_loan_or_request_on_document(
        patron_pid, document_pid
    )
    if loan_or_request_found:
        if (
            loan_or_request_found.state
            in current_app.config["CIRCULATION_STATES_LOAN_REQUEST"]
        ):
            raise PatronHasRequestOnDocumentError(patron_pid, document_pid)
        raise PatronHasLoanOnDocumentError(patron_pid, document_pid)

    _validate_delivery(kwargs.get("delivery"))

    transaction_user_pid = transaction_user_pid or str(current_user.id)

    # create a new loan
    record_uuid = uuid.uuid4()
    new_loan = dict(
        patron_pid=patron_pid,
        transaction_location_pid=transaction_location_pid,
        transaction_user_pid=transaction_user_pid,
    )
    pid = ils_circulation_loan_pid_minter(record_uuid, data=new_loan)
    loan = loan_cls.create(data=new_loan, id_=record_uuid)

    params = deepcopy(loan)
    params.update(document_pid=document_pid, **kwargs)

    # trigger the transition to request
    loan = current_circulation.circulation.trigger(
        loan, **dict(params, trigger="request")
    )

    return pid, loan


def patron_has_active_loan_on_item(patron_pid, item_pid):
    """Return True if patron has a active Loan for given item."""
    search = search_by_patron_item_or_document(
        patron_pid=patron_pid,
        item_pid=item_pid,
        filter_states=current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"],
    )
    search_result = search.execute()
    return search_result.hits.total.value > 0


def _checkout_loan(
    item_pid,
    patron_pid,
    transaction_location_pid,
    trigger="checkout",
    transaction_user_pid=None,
    delivery=None,
    **kwargs,
):
    """Checkout a loan."""
    transaction_user_pid = transaction_user_pid or str(current_user.id)
    loan_cls = current_circulation.loan_record_cls
    # create a new loan
    record_uuid = uuid.uuid4()
    new_loan = dict(
        patron_pid=patron_pid,
        transaction_location_pid=transaction_location_pid,
        transaction_user_pid=transaction_user_pid,
    )

    if delivery:
        new_loan["delivery"] = delivery
    # check if there is an existing request
    loan = patron_has_request_on_document(patron_pid, kwargs.get("document_pid"))
    if loan:
        loan = loan_cls.get_record_by_pid(loan.pid)
        pid = IlsCirculationLoanIdProvider.get(loan["pid"]).pid
        loan.update(new_loan)
    else:
        pid = ils_circulation_loan_pid_minter(record_uuid, data=new_loan)
        loan = loan_cls.create(data=new_loan, id_=record_uuid)

    params = deepcopy(loan)
    params.update(item_pid=item_pid, **kwargs)

    loan = current_circulation.circulation.trigger(
        loan, **dict(params, trigger=trigger)
    )
    return pid, loan


def checkout_loan(
    item_pid,
    patron_pid,
    transaction_location_pid,
    transaction_user_pid=None,
    force=False,
    **kwargs,
):
    """Create a new loan and trigger the first transition to ITEM_ON_LOAN.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    :param patron_pid: the PID value of the patron
    :param transaction_location_pid: the PID value of the location where the
        checkout is performed
    :param transaction_user_pid: the PID value of the user that performed the
        checkout
    :param force: if True, ignore the current status of the item and do perform
        the checkout. If False, the checkout will fail when the item cannot
        circulate.
    """

    if patron_has_active_loan_on_item(patron_pid=patron_pid, item_pid=item_pid):
        raise PatronHasLoanOnItemError(patron_pid, item_pid)
    optional_delivery = kwargs.get("delivery")
    if optional_delivery:
        _validate_delivery(optional_delivery)

    if force:
        _set_item_to_can_circulate(item_pid)

    return _checkout_loan(
        item_pid,
        patron_pid,
        transaction_location_pid,
        transaction_user_pid=transaction_user_pid,
        **kwargs,
    )


def _ensure_item_loanable_via_self_checkout(item_pid):
    """Self-checkout: return loanable item or raise when not loanable.

    Implements the self-checkout rules to loan an item.
    """
    item = current_app_ils.item_record_cls.get_record_by_pid(item_pid)
    item_dict = item.replace_refs()

    if item_dict["status"] != "CAN_CIRCULATE":
        raise ItemCannotCirculateError()

    circulation_state = item_dict["circulation"].get("state")
    has_active_loan = (
        circulation_state and circulation_state in CIRCULATION_STATES_LOAN_ACTIVE
    )
    if has_active_loan:
        raise ItemHasActiveLoanError(loan_pid=item_dict["circulation"]["loan_pid"])

    document = current_app_ils.document_record_cls.get_record_by_pid(
        item_dict["document_pid"]
    )
    document_dict = document.replace_refs()
    if document_dict["circulation"].get("overbooked", False):
        raise DocumentOverbookedError(
            f"Cannot self-checkout the overbooked document {item_dict['document_pid']}"
        )

    return item


def self_checkout_get_item_by_barcode(barcode):
    """Search for an item by barcode.

    :param barcode: the barcode of the item to search for
    :return item: the item that was found, or raise in case of errors
    """
    item_search = current_app_ils.item_search_cls()
    items = item_search.search_by_barcode(barcode).execute()
    if items.hits.total.value == 0:
        raise ItemNotFoundError(barcode=barcode)
    if items.hits.total.value > 1:
        raise MultipleItemsBarcodeFoundError(barcode)

    item_pid = items.hits[0].pid
    item = _ensure_item_loanable_via_self_checkout(item_pid)
    return item_pid, item


def self_checkout(
    item_pid, patron_pid, transaction_location_pid, transaction_user_pid=None, **kwargs
):
    """Perform self-checkout.

    :param item_pid: a dict containing `value` and `type` fields to
        uniquely identify the item.
    :param patron_pid: the PID value of the patron
    :param transaction_location_pid: the PID value of the location where the
        checkout is performed
    :param transaction_user_pid: the PID value of the user that performed the
        checkout
    """
    _ensure_item_loanable_via_self_checkout(item_pid["value"])
    return _checkout_loan(
        item_pid,
        patron_pid,
        transaction_location_pid,
        transaction_user_pid=transaction_user_pid,
        trigger="self_checkout",
        delivery=dict(method="SELF-CHECKOUT"),
        **kwargs,
    )


def bulk_extend_loans(patron_pid, **kwargs):
    """Bulk extend qualified loans."""
    loan_class = current_circulation.loan_record_cls

    loans_search = get_all_expiring_or_overdue_loans_by_patron_pid(
        patron_pid=patron_pid
    )

    extended_loans = []
    not_extended_loans = []

    for loan in loans_search.scan():
        loan_record = loan_class.get_record_by_pid(loan["pid"])
        params = deepcopy(loan_record)
        try:
            extended_loan = current_circulation.circulation.trigger(
                loan_record,
                **dict(
                    params,
                    trigger="extend",
                    transition_kwargs=dict(send_notification=False),
                ),
            )
            extended_loans.append(extended_loan)
        except (CirculationException, InvalidLoanExtendError):
            # has to be re-fetched due to mutable object returned by transition
            loan_record = loan_class.get_record_by_pid(loan["pid"])
            not_extended_loans.append(loan_record)

    return extended_loans, not_extended_loans


def update_dates_loan(
    record,
    start_date=None,
    end_date=None,
    request_start_date=None,
    request_expire_date=None,
):
    """Updates the dates of a loan."""
    state = record["state"]
    is_active_or_completed = (
        state in CIRCULATION_STATES_LOAN_ACTIVE
        or state in CIRCULATION_STATES_LOAN_COMPLETED
    )

    data = copy(record)

    if is_active_or_completed:
        today = date.today().strftime("%Y-%m-%d")
        if request_start_date or request_expire_date:
            raise IlsException(
                description="Cannot modify request dates of "
                "an active or completed loan."
            )
        if start_date:
            if start_date > today:
                raise InvalidParameterError(
                    description="Start date cannot be in "
                    "the future for active loans."
                )
            data["start_date"] = start_date
        if end_date:
            data["end_date"] = end_date
        if data["end_date"] < data["start_date"]:
            raise InvalidParameterError(description="Negative date range.")
    else:  # Pending or cancelled
        if start_date or end_date:
            raise IlsException(
                description="Cannot modify dates of " "a pending or cancelled loan."
            )
        if request_start_date:
            data["request_start_date"] = request_start_date
        if request_expire_date:
            data["request_expire_date"] = request_expire_date
        if data["request_expire_date"] < data["request_start_date"]:
            raise InvalidParameterError(description="Negative date range.")

    record.update(data)
    record.commit()
    db.session.commit()
    current_circulation.loan_indexer().index(record)

    return record


def circulation_default_loan_duration_for_item(item):
    """Transform circulation restrictions to timedelta for the given item."""
    value = item.get("circulation_restriction", "NO_RESTRICTION")
    if value == "ONE_WEEK":
        return timedelta(weeks=1)
    elif value == "TWO_WEEKS":
        return timedelta(weeks=2)
    elif value == "THREE_WEEKS":
        return timedelta(weeks=3)
    else:
        # default: NO_RESTRICTION
        return timedelta(weeks=4)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Helper proxy to the state object."""

from flask import current_app
from werkzeug.local import LocalProxy

current_ils_ill = LocalProxy(lambda: current_app.extensions["invenio-ils-ill"])
"""Helper proxy to get the current ILS ILL extension."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test loans item permissions."""

import json
from copy import deepcopy
from datetime import timedelta

import arrow
from flask import url_for
from flask_principal import UserNeed
from invenio_access.permissions import Permission
from invenio_search import current_search

from invenio_app_ils.errors import (
    LoanSelfCheckoutDocumentOverbooked,
    LoanSelfCheckoutItemActiveLoan,
    LoanSelfCheckoutItemInvalidStatus,
)
from invenio_app_ils.items.api import Item
from invenio_app_ils.items.serializers import item
from invenio_app_ils.proxies import current_app_ils
from tests.helpers import user_login, user_logout

NEW_LOAN = {
    "item_pid": "CHANGE ME IN EACH TEST",
    "document_pid": "docid-1",
    "patron_pid": "3",
    "transaction_location_pid": "locid-1",
    "pickup_location_pid": "locid-1",
    "delivery": {"method": "PICKUP"},
}

NEW_LOAN_REQUEST = {
    "document_pid": "CHANGE ME IN EACH TEST",
    "patron_pid": "3",
    "transaction_location_pid": "locid-1",
    "pickup_location_pid": "locid-1",
    "delivery": {"method": "PICKUP"},
}


def test_anonymous_cannot_request_loan_on_item(client, json_headers, testdata):
    """Test that anonymous users cannot request a loan on a item."""
    url = url_for("invenio_app_ils_circulation.loan_checkout")
    res = client.post(url, headers=json_headers, data=json.dumps(NEW_LOAN))
    assert res.status_code == 401


def test_librarian_can_checkout_item_for_user(client, json_headers, users, testdata):
    """Test that librarian can checkout a loan on a item for a patron."""
    user_login(client, "librarian", users)
    url = url_for("invenio_app_ils_circulation.loan_checkout")
    params = deepcopy(NEW_LOAN)
    params["transaction_user_pid"] = str(users["librarian"].id)
    params["item_pid"] = dict(type="pitmid", value="itemid-60")
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["item_pid"] == params["item_pid"]
    assert loan["document_pid"] == params["document_pid"]
    assert loan["patron_pid"] == params["patron_pid"]


def test_force_checkout_specific_permissions(
    app, client, json_headers, users, testdata
):
    """Test that only allowed users can perform a force checkout."""
    default_factory = app.config["ILS_VIEWS_PERMISSIONS_FACTORY"]
    librarian2 = users["librarian2"]

    # override default permission factory to require specific permission for
    # force-checkout action
    def custom_views_permissions_factory(action):
        if action == "circulation-loan-force-checkout":
            # fake permission for a specific user
            return Permission(UserNeed(librarian2.id))
        else:
            return default_factory(action)

    app.config["ILS_VIEWS_PERMISSIONS_FACTORY"] = custom_views_permissions_factory

    # prepare request
    url = url_for("invenio_app_ils_circulation.loan_checkout")
    params = deepcopy(NEW_LOAN)
    params["force"] = True
    params["item_pid"] = dict(type="pitmid", value="itemid-MISSING")
    params["transaction_user_pid"] = str(librarian2.id)

    # force-checkout as librarian should fail
    user_login(client, "librarian", users)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 403

    # force-checkout as librarian2 should succeed
    user_login(client, "librarian2", users)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["item_pid"] == params["item_pid"]
    assert loan["document_pid"] == params["document_pid"]
    assert loan["patron_pid"] == params["patron_pid"]

    # restore default config
    app.config["ILS_VIEWS_PERMISSIONS_FACTORY"] = default_factory


def test_checkout_conditions_librarian(client, json_headers, users, testdata):
    """Test checkout conditions user."""
    librarian = users["librarian"]
    user_login(client, "librarian", users)
    url = url_for("invenio_app_ils_circulation.loan_checkout")

    params = deepcopy(NEW_LOAN)
    params["item_pid"] = dict(type="pitmid", value="itemid-61")
    params["transaction_user_pid"] = str(librarian.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202

    params = deepcopy(NEW_LOAN)
    params["item_pid"] = dict(type="pitmid", value="itemid-MISSING")
    params["transaction_user_pid"] = str(librarian.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    # missing
    assert res.status_code == 400

    params = deepcopy(NEW_LOAN)
    params["force"] = True
    params["item_pid"] = dict(type="pitmid", value="itemid-MISSING")
    params["transaction_user_pid"] = str(librarian.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    # missing but force
    assert res.status_code == 202
    item = Item.get_record_by_pid(params["item_pid"]["value"])
    assert item["status"] == "CAN_CIRCULATE"


def test_checkout_loader_start_end_dates(app, client, json_headers, users, testdata):
    """Test that start and end dates request parameters."""
    librarian = users["librarian"]
    user_login(client, "librarian", users)
    item_pid = dict(type="pitmid", value="itemid-61")
    url = url_for("invenio_app_ils_circulation.loan_checkout")
    # fake duration for first item
    loan_duration_timedelta = app.config["CIRCULATION_POLICIES"]["checkout"][
        "duration_default"
    ](dict(item_pid=item_pid), dict())
    now = arrow.utcnow()

    # it should succeed when no start/end dates provided
    start_date = now.date().isoformat()
    end_date = (now + loan_duration_timedelta).date().isoformat()

    params = deepcopy(NEW_LOAN)
    params["item_pid"] = item_pid
    params["transaction_user_pid"] = str(librarian.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["item_pid"] == params["item_pid"]
    assert loan["start_date"] == start_date
    assert loan["end_date"] == end_date
    assert loan["transaction_date"]

    # it should succeed when no end date provided
    _start_date = now + timedelta(days=3)
    start_date = _start_date.date().isoformat()
    end_date = (_start_date + loan_duration_timedelta).date().isoformat()

    params = deepcopy(NEW_LOAN)
    params["item_pid"] = dict(type="pitmid", value="itemid-62")
    params["start_date"] = start_date
    params["transaction_user_pid"] = str(librarian.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["item_pid"] == params["item_pid"]
    assert loan["start_date"] == start_date
    assert loan["end_date"] == end_date
    assert loan["transaction_date"]

    start_date = (now + timedelta(days=3)).date().isoformat()
    end_date = (now + timedelta(days=15)).date().isoformat()

    # it should succeed when start/end dates provided
    params = deepcopy(NEW_LOAN)
    params["item_pid"] = dict(type="pitmid", value="itemid-63")
    params["start_date"] = start_date
    params["end_date"] = end_date
    params["transaction_user_pid"] = str(librarian.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "ITEM_ON_LOAN"
    assert loan["item_pid"] == params["item_pid"]
    assert loan["start_date"] == start_date
    assert loan["end_date"] == end_date
    assert loan["transaction_date"]

    # it should fail when only end date provided
    params = deepcopy(NEW_LOAN)
    params["item_pid"] = dict(type="pitmid", value="itemid-63")
    params["end_date"] = end_date
    params["transaction_user_pid"] = str(librarian.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400

    # it should fail when wrong date provided
    start_date = (now + timedelta(days=3)).date().isoformat()
    past_end_date = now - timedelta(days=30)
    params = deepcopy(NEW_LOAN)
    params["item_pid"] = dict(type="pitmid", value="itemid-63")
    params["start_date"] = start_date
    params["end_date"] = past_end_date.date().isoformat()
    params["transaction_user_pid"] = str(librarian.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400


def test_self_checkout_search(app, client, json_headers, users, testdata):
    """Test self-checkout search."""
    app.config["ILS_SELF_CHECKOUT_ENABLED"] = True

    # test that anonymous user cannot search for barcodes
    url = url_for("invenio_app_ils_circulation.loan_self_checkout")
    res = client.get(f"{url}?barcode=123456", headers=json_headers)
    assert res.status_code == 401

    user_login(client, "patron2", users)

    # test the missing parameter error
    url = url_for("invenio_app_ils_circulation.loan_self_checkout")
    res = client.get(url, headers=json_headers)
    assert res.status_code == 400

    # test that authenticated user can search for barcodes, but it will return
    # 400 when not found
    unexisting_barcode = "123456"
    res = client.get(f"{url}?barcode={unexisting_barcode}", headers=json_headers)
    assert res.status_code == 400

    # test that an error is returned when the item cannot circulate
    missing_item_barcode = "123456789-1"
    url = url_for("invenio_app_ils_circulation.loan_self_checkout")
    res = client.get(f"{url}?barcode={missing_item_barcode}", headers=json_headers)
    assert res.status_code == 400
    # assert that the payload will contain the key error with a msg
    response = res.get_json()
    assert LoanSelfCheckoutItemInvalidStatus.description in response["message"]
    assert LoanSelfCheckoutItemInvalidStatus.supportCode in response["supportCode"]

    # create a loan on the same patron, and another one on another patron
    user_login(client, "librarian", users)
    url = url_for("invenio_app_ils_circulation.loan_checkout")

    for item_pid, patron_pid in [
        ("itemid-60", "2"),  # barcode 123456789-60
        ("itemid-61", "1"),  # barcode 123456789-61
    ]:
        params = deepcopy(NEW_LOAN)
        params["transaction_user_pid"] = str(users["librarian"].id)
        params["item_pid"] = dict(type="pitmid", value=item_pid)
        params["patron_pid"] = patron_pid
        res = client.post(url, headers=json_headers, data=json.dumps(params))
        assert res.status_code == 202

    # ensure new loans and related items are fully indexed
    current_search.flush_and_refresh(index="*")

    user_login(client, "patron2", users)

    # test that an error is returned when the item is already on loan by the same user
    on_loan_same_patron_barcode = "123456789-60"
    url = url_for("invenio_app_ils_circulation.loan_self_checkout")
    res = client.get(
        f"{url}?barcode={on_loan_same_patron_barcode}", headers=json_headers
    )
    assert res.status_code == 400
    # assert that the payload will contain the key error with a msg
    response = res.get_json()
    assert LoanSelfCheckoutItemActiveLoan.description in response["message"]
    assert LoanSelfCheckoutItemActiveLoan.supportCode in response["supportCode"]

    # test that an error is returned when the item is already on loan by another user
    on_loan_other_patron_barcode = "123456789-61"
    url = url_for("invenio_app_ils_circulation.loan_self_checkout")
    res = client.get(
        f"{url}?barcode={on_loan_other_patron_barcode}", headers=json_headers
    )
    assert res.status_code == 400
    # assert that the payload will contain the key error with a msg
    response = res.get_json()
    assert LoanSelfCheckoutItemActiveLoan.description in response["message"]
    assert LoanSelfCheckoutItemActiveLoan.supportCode in response["supportCode"]

    # test happy path
    available_barcode = "123456789-10"
    url = url_for("invenio_app_ils_circulation.loan_self_checkout")
    res = client.get(f"{url}?barcode={available_barcode}", headers=json_headers)
    assert res.status_code == 200
    response = res.get_json()
    assert response["metadata"]["pid"] == "itemid-10"


def test_self_checkout(app, client, json_headers, users, testdata):
    """Test self-checkout."""

    def _create_request(patron, document_pid):
        url = url_for("invenio_app_ils_circulation.loan_request")
        user = user_login(client, patron, users)
        params = deepcopy(NEW_LOAN_REQUEST)
        params["document_pid"] = document_pid
        params["patron_pid"] = str(user.id)
        params["transaction_user_pid"] = str(user.id)
        res = client.post(url, headers=json_headers, data=json.dumps(params))
        assert res.status_code == 202
        current_search.flush_and_refresh(index="*")

    def _self_checkout(patron, item_pid, document_pid):
        params = deepcopy(NEW_LOAN)
        params["document_pid"] = document_pid
        params["item_pid"] = dict(type="pitmid", value=item_pid)
        params["patron_pid"] = str(patron.id)
        params["transaction_user_pid"] = str(patron.id)
        return client.post(url, headers=json_headers, data=json.dumps(params))

    url = url_for("invenio_app_ils_circulation.loan_self_checkout")

    app.config["ILS_SELF_CHECKOUT_ENABLED"] = False

    # test a logged in user cannot self-checkout when the feature is disabled
    patron2 = user_login(client, "patron2", users)
    res = client.post(url, headers=json_headers)
    assert res.status_code == 403

    user_logout(client)
    app.config["ILS_SELF_CHECKOUT_ENABLED"] = True

    # test that anonymous user cannot self-checkout
    res = client.post(url, headers=json_headers)
    assert res.status_code == 401

    # test overbooked books
    # create multiple requests from different patrons
    _create_request("patron1", "docid-15")
    _create_request("patron3", "docid-15")

    document_rec = current_app_ils.document_record_cls.get_record_by_pid("docid-15")
    document = document_rec.replace_refs()
    assert document["circulation"]["overbooked"]

    # test that user cannot self-checkout an overbooked book, without having a request
    patron2 = user_login(client, "patron2", users)
    res = _self_checkout(patron2, "itemid-71", "docid-15")
    assert res.status_code == 400
    response = res.get_json()
    assert LoanSelfCheckoutDocumentOverbooked.description in response["message"]
    assert LoanSelfCheckoutDocumentOverbooked.supportCode in response["supportCode"]

    # test that user cannot self-checkout an overbooked book, having a request
    # create request from the same patron
    _create_request("patron2", "docid-15")
    patron2 = user_login(client, "patron2", users)
    res = _self_checkout(patron2, "itemid-71", "docid-15")
    assert res.status_code == 400
    response = res.get_json()
    assert LoanSelfCheckoutDocumentOverbooked.description in response["message"]
    assert LoanSelfCheckoutDocumentOverbooked.supportCode in response["supportCode"]

    # test that user can self-checkout having a prior request
    _create_request("patron2", "docid-16")

    patron2 = user_login(client, "patron2", users)
    res = _self_checkout(patron2, "itemid-72", "docid-16")
    assert res.status_code == 202
    response = res.get_json()
    assert response["metadata"]["delivery"]["method"] == "SELF-CHECKOUT"

    # test that user can self-checkout without having a prior request, even if there
    # are other requests on the book (but not overbooked)
    _create_request("patron1", "docid-17")

    patron2 = user_login(client, "patron2", users)
    res = _self_checkout(patron2, "itemid-73", "docid-17")
    assert res.status_code == 202
    response = res.get_json()
    assert response["metadata"]["delivery"]["method"] == "SELF-CHECKOUT"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""APIs for CRUD operations around Records Relations."""

from copy import deepcopy

from invenio_app_ils.errors import RecordHasReferencesError, RecordRelationsError
from invenio_app_ils.records.api import IlsRecord
from invenio_app_ils.relations.api import (
    MULTIPART_MONOGRAPH_RELATION,
    PARENT_CHILD_RELATION_TYPES,
    SEQUENCE_RELATION_TYPES,
    SERIAL_RELATION,
    SIBLINGS_RELATION_TYPES,
    ParentChildRelation,
    SequenceRelation,
    SiblingsRelation,
)


class RecordRelationsExtraMetadata(object):
    """Utilities to manage the extra metadata field."""

    _field_name = "relations_extra_metadata"

    @classmethod
    def field_name(cls):
        """Return field name property."""
        return cls._field_name

    @classmethod
    def build_metadata_object(cls, pid_value, pid_type, **kwargs):
        """Build and return the payload to be added to extra metadata field."""
        r = {"pid_value": pid_value, "pid_type": pid_type}
        r.update(kwargs)
        return r

    @classmethod
    def get_extra_metadata_from(cls, record, relation_name, pid_value, pid_type):
        """Return the extra metadata dict for the given PID and type."""
        metadata = record.get(cls.field_name(), {}).get(relation_name, [])
        for m in metadata:
            has_same_pid = (
                m.get("pid_value", "") == pid_value
                and m.get("pid_type", "") == pid_type
            )
            if has_same_pid:
                return deepcopy(m)
        return {}

    @classmethod
    def add_extra_metadata_to(
        cls, record, relation_name, pid_value, pid_type, **kwargs
    ):
        """Add a new extra metadata dict for the given PID and type."""
        metadata = record.setdefault(cls.field_name(), {})
        relation_metadata = metadata.setdefault(relation_name, [])
        for m in relation_metadata:
            if (
                m.get("pid_value", "") == pid_value
                and m.get("pid_type", "") == pid_type
            ):
                raise RecordRelationsError(
                    "The record PID `{}` has already metadata for the relation"
                    " `{}` and record PID `{}`".format(
                        record.pid.pid_value, relation_name, pid_value
                    )
                )
        obj = RecordRelationsExtraMetadata.build_metadata_object(
            pid_value, pid_type, **kwargs
        )
        relation_metadata.append(obj)
        record.commit()

    @classmethod
    def remove_extra_metadata_from(cls, record, relation_name, pid_value, pid_type):
        """Remove any presence of the given PID in extra metadata."""
        field = cls.field_name()
        if field in record and relation_name in record[field]:

            def keep_pid_func(m):
                """Keep the pid."""
                return not (
                    m.get("pid_value", "") == pid_value
                    and m.get("pid_type", "") == pid_type
                )

            remaining_relations = list(
                filter(keep_pid_func, record[field][relation_name])
            )

            if remaining_relations != record[field][relation_name]:
                # if there are no more relations of this type, remove the obj
                if not remaining_relations:
                    del record[field][relation_name]
                else:
                    record[field][relation_name] = remaining_relations

                # if there are 0 extra metadata left, delete the field
                if not record[field]:
                    del record[field]

                record.commit()


class RecordRelations(object):
    """Record relations object."""

    relation_types = []

    def _validate_relation_type(self, relation_type):
        """Validate the given relation type."""
        if relation_type not in self.relation_types:
            rel_names = ",".join([rt.name for rt in self.relation_types])
            raise RecordRelationsError(
                "Relation type must be one of `{}`".format(rel_names)
            )

    def _validate_relation_between_records(self, rec1, rec2, relation_name):
        """Abstract method to validate relation between records."""
        raise NotImplementedError

    def add(self, rec1, rec2, relation_type, **kwargs):
        """Add a new relation between the given records."""
        raise NotImplementedError

    def remove(self, rec1, rec2, relation_type):
        """Remove an existing relation between the given records."""
        raise NotImplementedError


class RecordRelationsParentChild(RecordRelations):
    """Add/Remove operations for Parent-Child relations."""

    allowed_metadata = ["volume"]

    def __init__(self):
        """Constructor."""
        self.relation_types = PARENT_CHILD_RELATION_TYPES

    def _validate_relation_between_records(self, parent, child, relation_type):
        """Validate relation between type of records."""
        from invenio_app_ils.documents.api import Document
        from invenio_app_ils.series.api import Series

        if relation_type.name == "multipart_monograph":
            pcr = ParentChildRelation(relation_type)
            relations = pcr.get_relations_by_child(child.pid)
            if len(relations) > 0:
                raise RecordRelationsError(
                    "Cannot create a relation `{}` between PID `{}` as parent"
                    " and PID `{}` as child. Record `{}` has already a"
                    " multipart monograph.".format(
                        relation_type.name,
                        parent.pid.pid_value,
                        child.pid.pid_value,
                        child.pid.pid_value,
                    )
                )

        # when child is Document, parent is any type of Series
        is_series_doc = isinstance(child, Document) and isinstance(parent, Series)
        # when child is Multipart Monograph, parent is only Serials
        is_serial_mm = (
            isinstance(child, Series)
            and isinstance(parent, Series)
            and child["mode_of_issuance"] == "MULTIPART_MONOGRAPH"
            and parent["mode_of_issuance"] == "SERIAL"
        )

        if not (is_series_doc or is_serial_mm):
            raise RecordRelationsError(
                "Cannot create a relation `{}` between PID `{}` as parent and "
                "PID `{}` as child.".format(
                    relation_type.name,
                    parent.pid.pid_value,
                    child.pid.pid_value,
                )
            )
        return True

    def add(self, parent, child, relation_type, **kwargs):
        """Add a new relation between given parent and child records."""
        self._validate_relation_type(relation_type)
        self._validate_relation_between_records(
            parent=parent, child=child, relation_type=relation_type
        )

        pcr = ParentChildRelation(relation_type)
        pcr.add(parent_pid=parent.pid, child_pid=child.pid)

        # relation metadata is allowed only for MULTIPART_MONOGRAPH
        relation_allows_metadata = relation_type in (
            MULTIPART_MONOGRAPH_RELATION,
            SERIAL_RELATION,
        )
        # check for allowed relation metadata (e.g. `volume`)
        has_allowed_metadata = any(
            [kwargs.get(metadata) for metadata in self.allowed_metadata]
        )

        if relation_allows_metadata and has_allowed_metadata:
            # filter and keep only allowed metadata
            allowed = {k: v for k, v in kwargs.items() if k in self.allowed_metadata}
            # store relation metadata in the child record
            RecordRelationsExtraMetadata.add_extra_metadata_to(
                child,
                relation_type.name,
                parent.pid.pid_value,
                parent.pid.pid_type,
                **allowed,
            )

        # return the allegedly modified record
        return child

    def remove(self, parent, child, relation_type):
        """Remove a relation between given parent and child records."""
        self._validate_relation_type(relation_type)
        pcr = ParentChildRelation(relation_type)
        pcr.remove(parent_pid=parent.pid, child_pid=child.pid)

        # remove any metadata for this relation, if any
        RecordRelationsExtraMetadata.remove_extra_metadata_from(
            child,
            relation_type.name,
            parent.pid.pid_value,
            parent.pid.pid_type,
        )
        return child


class RecordRelationsSiblings(RecordRelations):
    """Add/Remove operations for Siblings relations."""

    allowed_metadata = ["note"]

    def __init__(self):
        """Constructor."""
        self.relation_types = SIBLINGS_RELATION_TYPES

    def _validate_relation_between_records(self, first, second, relation_name):
        """Validate relation between type of records."""
        from invenio_app_ils.documents.api import Document
        from invenio_app_ils.series.api import Series

        # records must be of the same type
        same_document = isinstance(first, Document) and isinstance(second, Document)
        same_series = (
            isinstance(first, Series)
            and isinstance(second, Series)
            and first["mode_of_issuance"] == second["mode_of_issuance"]
        )
        valid_edition_relation = relation_name == "edition" and (
            (
                isinstance(first, Document)
                and isinstance(second, Series)
                and second["mode_of_issuance"] == "MULTIPART_MONOGRAPH"
            )
            or (
                isinstance(second, Document)
                and isinstance(first, Series)
                and first["mode_of_issuance"] == "MULTIPART_MONOGRAPH"
            )
        )

        valid_edition_fields = (
            relation_name == "edition"
            and first.get("edition", False)
            and second.get("edition", False)
        )

        valid_language_fields = (
            relation_name == "language"
            and first.get("languages", False)
            and second.get("languages", False)
        )

        valid_other_fields = relation_name == "other"

        equal_editions = relation_name == "edition" and first.get(
            "edition"
        ) == second.get("edition")

        equal_languages = relation_name == "language" and first.get(
            "languages"
        ) == second.get("languages")

        if not (same_document or same_series or valid_edition_relation):
            raise RecordRelationsError(
                "Cannot create relation `{}` between PID `{}` and  PID `{}`,"
                " they are different record types".format(
                    relation_name, first.pid.pid_value, second.pid.pid_value
                )
            )

        if not (valid_edition_fields or valid_language_fields or valid_other_fields):
            raise RecordRelationsError(
                "Cannot create relation `{}` "
                "between PID `{}` and  PID `{}`,"
                " one of the records is missing {} fields".format(
                    relation_name,
                    first.pid.pid_value,
                    second.pid.pid_value,
                    relation_name,
                )
            )

        if equal_editions or equal_languages:
            raise RecordRelationsError(
                "Cannot create relation `{}` "
                "between PID `{}` and  PID `{}`,"
                " records have equal {} fields".format(
                    relation_name,
                    first.pid.pid_value,
                    second.pid.pid_value,
                    relation_name,
                )
            )

        return True

    def add(self, first, second, relation_type, **kwargs):
        """Add a new relation between given first and second records."""
        self._validate_relation_type(relation_type)
        self._validate_relation_between_records(
            first=first, second=second, relation_name=relation_type.name
        )

        sr = SiblingsRelation(relation_type)
        sr.add(first_pid=first.pid, second_pid=second.pid)

        # check for allowed relation metadata (e.g. `note`)
        has_allowed_metadata = any(
            [kwargs.get(metadata) for metadata in self.allowed_metadata]
        )

        if has_allowed_metadata:
            # filter and keep only allowed metadata
            allowed = {k: v for k, v in kwargs.items() if k in self.allowed_metadata}
            # store relation metadata in the first record
            RecordRelationsExtraMetadata.add_extra_metadata_to(
                first,
                relation_type.name,
                second.pid.pid_value,
                second.pid.pid_type,
                **allowed,
            )
        return first

    def remove(self, first, second, relation_type):
        """Remove the relation between the first and the second."""
        self._validate_relation_type(relation_type)
        sr = SiblingsRelation(relation_type)
        sr.remove(pid=second.pid)

        # remove any metadata for this relation, if any
        # both first and second could have metadata for the relation
        RecordRelationsExtraMetadata.remove_extra_metadata_from(
            first, relation_type.name, second.pid.pid_value, second._pid_type
        )
        RecordRelationsExtraMetadata.remove_extra_metadata_from(
            second, relation_type.name, first.pid.pid_value, first._pid_type
        )
        return first, second


class RecordRelationsSequence(RecordRelations):
    """Add/Remove operations for Sequence relations."""

    def __init__(self):
        """Constructor."""
        self.relation_types = SEQUENCE_RELATION_TYPES

    def _validate_relation_between_records(self, previous_rec, next_rec, relation_name):
        """Validate relation between type of records."""
        from invenio_app_ils.series.api import Series

        # records must be of the same type, Sequences support only Series
        allowed_types = [Series]

        for record_type in allowed_types:
            if isinstance(previous_rec, record_type) and isinstance(
                next_rec, record_type
            ):
                return True

        raise RecordRelationsError(
            "Cannot create a relation `{}` between PID `{}` with type {} "
            " and PID `{}` with type {}.".format(
                relation_name,
                previous_rec.pid.pid_value,
                previous_rec.pid.pid_type,
                next_rec.pid.pid_value,
                next_rec.pid.pid_type,
            )
        )

    def add(self, previous_rec, next_rec, relation_type, **kwargs):
        """Add a new sequence relation between previous and next records."""
        self._validate_relation_type(relation_type)
        self._validate_relation_between_records(
            previous_rec, next_rec, relation_type.name
        )

        sequence_relation = SequenceRelation(relation_type)
        sequence_relation.add(previous_pid=previous_rec.pid, next_pid=next_rec.pid)

    def remove(self, previous_rec, next_rec, relation_type):
        """Remove sequence relation between previous and next records."""
        self._validate_relation_type(relation_type)
        sequence_relation = SequenceRelation(relation_type)
        sequence_relation.remove(previous_pid=previous_rec.pid, next_pid=next_rec.pid)


class IlsRecordWithRelations(IlsRecord):
    """Add relations functionalities to records."""

    @property
    def relations(self):
        """Get record relations."""
        from .retriever import get_relations

        return get_relations(self)

    def clear(self):
        """Clear IlsRecordWithRelations record."""
        extra_metadata_field_name = RecordRelationsExtraMetadata.field_name()
        extra_metadata = self.get(extra_metadata_field_name, {})
        super().clear()
        self[extra_metadata_field_name] = extra_metadata

    def delete(self, **kwargs):
        """Delete record with relations."""
        related_refs = set()
        for name, related_objects in self.relations.items():
            for obj in related_objects:
                related_refs.add("{pid_value}:{pid_type}".format(**obj))
        if related_refs:
            raise RecordHasReferencesError(
                record_type=self.__class__.__name__,
                record_id=self["pid"],
                ref_type="related",
                ref_ids=sorted(ref for ref in related_refs),
            )
        super().delete(**kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Vocabularies API module."""

from functools import partial

from flask import current_app
from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2
from invenio_records_rest.utils import obj_or_import_string
from invenio_search import current_search

from invenio_app_ils.fetchers import pid_fetcher
from invenio_app_ils.proxies import current_app_ils

# hardcoded values for vocabularies that are not customizable via JSON files
VOCABULARY_TYPE_LANGUAGE = "language"
VOCABULARY_TYPE_COUNTRY = "country"
VOCABULARY_TYPE_LICENSE = "license"

VOCABULARY_PID_TYPE = "vocid"
VOCABULARY_PID_MINTER = "vocid"
VOCABULARY_PID_FETCHER = "vocid"

VocabularyIdProvider = type(
    "VocabularyIdProvider",
    (RecordIdProviderV2,),
    dict(pid_type=VOCABULARY_PID_TYPE, default_status=PIDStatus.REGISTERED),
)
vocabulary_pid_minter = None
vocabulary_pid_fetcher = partial(
    pid_fetcher, provider_cls=VocabularyIdProvider, pid_field="id"
)


class Vocabulary(dict):
    """Vocabulary record class."""

    _index = "vocabularies-vocabulary-v1.0.0"
    _doc_type = None
    _schema = "vocabularies/vocabulary-v1.0.0.json"

    def __init__(self, type, key, text, data=None):
        """Create a `Vocabulary` instance.

        Vocabulary instances are not stored in the database
        but are indexed in ElasticSearch.
        """
        self.type = type
        self.key = key
        self.text = text
        self.data = data or {}

        # Needed by the indexer
        self.id = "{}-{}".format(type, key).lower()
        self.revision_id = 1

    def dumps(self):
        """Return python representation of vocabulary metadata."""
        return {
            "id": self.id,
            "type": self.type,
            "key": self.key,
            "text": self.text,
            "data": self.data,
        }

    def __repr__(self):
        """Representation of a vocabulary."""
        return "Vocabulary(id={}, key={}, type={}, text={!r}, data={})".format(
            self.id, self.key, self.type, self.text, self.data
        )


def validate_vocabulary(f):
    """Decorator to validate vocabulary schema."""

    def inner(self):
        vocabularies = f(self)
        for vocabulary in vocabularies:
            self.validate(vocabulary)
        return vocabularies

    return inner


def load_vocabularies(source_name, *args, **kwargs):
    """Load vocabularies from a vocabulary source."""
    source_name = current_app.config["ILS_VOCABULARY_SOURCES"][source_name]
    source = obj_or_import_string(source_name)(*args, **kwargs)
    return source.load()


def delete_vocabulary_from_index(type, force=False, key=None):
    """Delete vocabulary from index given a type and optionally a key."""
    search = current_app_ils.vocabulary_search_cls()
    if key is None:
        results = search.search_by_type(type)
    else:
        results = search.search_by_type_and_key(type, key)

    count = results.count()
    if force:
        results.delete()
        current_search.flush_and_refresh(index=search.Meta.index)

    return count



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Schema for Borrowing Request create loan action for marshmallow loader."""

import arrow
from invenio_circulation.records.loaders.schemas.json import DateString
from invenio_rest.serializer import BaseSchema as InvenioBaseSchema
from marshmallow import EXCLUDE, ValidationError, fields, post_load, pre_load, validates

from invenio_app_ils.circulation.loaders.schemas.json.base import (
    transaction_location_pid_validator,
)


class CreateLoanSchemaV1(InvenioBaseSchema):
    """Schema for Borrowing Request create loan action."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    loan_start_date = DateString(required=True)
    loan_end_date = DateString(required=True)
    transaction_location_pid = fields.Str(required=True)

    @pre_load
    def validate_statuses(self, data, **kwargs):
        """Validate status and that the loan does not exist yet."""
        record = self.context["record"]
        if record["status"] != "REQUESTED":
            raise ValidationError(
                "A loan can be created only when the borrowing request is in "
                "requested status."
            )

        loan_pid = record.get("patron_loan", {}).get("pid")
        if loan_pid:
            raise ValidationError(
                "This borrowing request {} has already a loan ({}).".format(
                    record["pid"], loan_pid
                )
            )
        return data

    @validates("loan_start_date")
    def validate_loan_start_date(self, value, **kwargs):
        """Validate loan_start_date field."""
        if arrow.get(value).date() < arrow.now().date():
            raise ValidationError(
                "The loan start date cannot be in the past.",
                "loan_start_date",
            )

    @validates("loan_end_date")
    def validate_loan_end_date(self, value, **kwargs):
        """Validate loan_end_date field."""
        if arrow.get(value).date() < arrow.now().date():
            raise ValidationError(
                "The loan end date cannot be in the past.",
                "loan_end_date",
            )

    @validates("transaction_location_pid")
    def validate_transaction_location_pid(self, value, **kwargs):
        """Validate transaction_location_pid field."""
        transaction_location_pid_validator(value)

    @post_load
    def postload_checks(self, data, **kwargs):
        """Validate dates values."""
        start = arrow.get(data["loan_start_date"]).date()
        end = arrow.get(data["loan_end_date"]).date()
        if end < start:
            raise ValidationError(
                {
                    "loan_start_date": [
                        "The loan start date cannot be after the end date."
                    ],
                    "loan_end_date": [
                        "The loan end date cannot be before the start date."
                    ],
                }
            )

        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test loans default durations."""

from datetime import timedelta

import arrow

from invenio_app_ils.ill.api import (
    circulation_default_extension_duration,
    circulation_default_loan_duration,
)

FAKE_LOAN_ITEM_NO_RESTRICTIONS = {"item_pid": {"type": "pitmid", "value": "itemid-1"}}
FAKE_LOAN_ITEM_ONE_WEEK = {"item_pid": {"type": "pitmid", "value": "itemid-2"}}
FAKE_LOAN_ITEM_TWO_WEEKS = {"item_pid": {"type": "pitmid", "value": "itemid-3"}}
FAKE_LOAN_ITEM_THREE_WEEKS = {"item_pid": {"type": "pitmid", "value": "itemid-4"}}


def test_loans_default_durations(testdata):
    """Test loans default durations."""
    tomorrow = arrow.utcnow() + timedelta(days=1)
    not_overdue_end_date = tomorrow.date().isoformat()
    for duration_func in (
        circulation_default_loan_duration,
        circulation_default_extension_duration,
    ):
        FAKE_LOAN_ITEM_NO_RESTRICTIONS["end_date"] = not_overdue_end_date
        assert duration_func(FAKE_LOAN_ITEM_NO_RESTRICTIONS, None) == timedelta(weeks=4)

        FAKE_LOAN_ITEM_ONE_WEEK["end_date"] = not_overdue_end_date
        assert duration_func(FAKE_LOAN_ITEM_ONE_WEEK, None) == timedelta(weeks=1)

        FAKE_LOAN_ITEM_TWO_WEEKS["end_date"] = not_overdue_end_date
        assert duration_func(FAKE_LOAN_ITEM_TWO_WEEKS, None) == timedelta(weeks=2)

        FAKE_LOAN_ITEM_THREE_WEEKS["end_date"] = not_overdue_end_date
        assert duration_func(FAKE_LOAN_ITEM_THREE_WEEKS, None) == timedelta(weeks=3)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for document resolvers."""

from invenio_app_ils.documents.api import Document


def test_document_resolvers(app, testdata):
    """Test item resolving from loan."""
    doc_pid = testdata["documents"][0]["pid"]
    document = Document.get_record_by_pid(doc_pid)
    document = document.replace_refs()

    # relations
    assert "relations" in document

    # circulation
    assert "active_loans_count" in document["circulation"]
    assert "pending_loans_count" in document["circulation"]
    assert "overdue_loans_count" in document["circulation"]
    assert "past_loans_count" in document["circulation"]

    # item and eitems
    assert document["items"]["total"] == 9 and document["items"]["hits"]
    assert document["eitems"]["total"] == 3 and document["eitems"]["hits"]

    # stock
    mediums = set([item["medium"] for item in document["items"]["hits"]])
    mediums.add("E-BOOK")
    assert set(document["stock"]["mediums"]) == mediums



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolve the files for the EItem."""

import jsonresolver
from werkzeug.routing import Rule

from invenio_app_ils.eitems.api import EItem

# Note: there must be only one resolver per file,
# otherwise only the last one is registered


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the referred Document for an EItem record."""
    from flask import current_app

    def files_resolver(eitem_pid):
        """Return the Document record for the given EItem or raise."""
        files = []
        for obj in EItem.get_record_by_pid(eitem_pid).files:
            files.append(
                {
                    "bucket": obj.bucket_id,
                    "checksum": obj.file.checksum,
                    "file_id": obj.file.id,
                    "key": obj.key,
                    "size": obj.file.size,
                    "version_id": str(obj.version_id),
                }
            )
        return files

    url_map.add(
        Rule(
            "/api/resolver/eitems/<eitem_pid>/files",
            endpoint=files_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Common pytest fixtures and plugins."""

import os

import jinja2
import pytest
from flask import Blueprint
from flask_webpackext.manifest import (
    JinjaManifest,
    JinjaManifestEntry,
    JinjaManifestLoader,
)
from invenio_access.models import ActionRoles
from invenio_access.permissions import superuser_access
from invenio_accounts.models import Role
from invenio_app.factory import create_app as _create_app
from invenio_userprofiles import UserProfile

from invenio_app_ils.permissions import backoffice_access_action


#
# Mock the webpack manifest to avoid having to compile the full assets.
#
class MockJinjaManifest(JinjaManifest):
    """Mock manifest."""

    def __getitem__(self, key):
        """Get a manifest entry."""
        return JinjaManifestEntry(key, [key])

    def __getattr__(self, name):
        """Get a manifest entry."""
        return JinjaManifestEntry(name, [name])


class MockManifestLoader(JinjaManifestLoader):
    """Manifest loader creating a mocked manifest."""

    def load(self, filepath):
        """Load the manifest."""
        return MockJinjaManifest()


@pytest.fixture(scope="module")
def app_config(app_config):
    """Create test app."""
    app_config["WEBPACKEXT_MANIFEST_LOADER"] = MockManifestLoader
    return app_config


@pytest.fixture()
def users(app, db):
    """Create admin, librarians and patrons."""
    # with Postgresql, when dropping the User table, the sequence is not
    # automatically reset to 1, causing issues with the tests demo data.
    db.session.execute("ALTER SEQUENCE IF EXISTS accounts_user_id_seq RESTART")
    db.session.commit()

    with db.session.begin_nested():
        datastore = app.extensions["security"].datastore
        # create users
        patron1 = datastore.create_user(
            email="patron1@test.com", password="123456", active=True
        )

        patron2 = datastore.create_user(
            email="patron2@test.com", password="123456", active=True
        )
        patron3 = datastore.create_user(
            email="patron3@test.com", password="123456", active=True
        )
        librarian = datastore.create_user(
            email="librarian@test.com", password="123456", active=True
        )
        librarian2 = datastore.create_user(
            email="librarian2@test.com", password="123456", active=True
        )
        admin = datastore.create_user(
            email="admin@test.com", password="123456", active=True
        )
        # Give role to admin
        admin_role = Role(name="admin")
        db.session.add(ActionRoles(action=superuser_access.value, role=admin_role))
        datastore.add_role_to_user(admin, admin_role)
        # Give role to librarian
        librarian_role = Role(name="librarian")
        db.session.add(
            ActionRoles(action=backoffice_access_action.value, role=librarian_role)
        )
        datastore.add_role_to_user(librarian, librarian_role)
        # Give role to librarian2
        db.session.add(
            ActionRoles(action=backoffice_access_action.value, role=librarian_role)
        )
        datastore.add_role_to_user(librarian2, librarian_role)
    db.session.commit()

    for patron, name in [
        (admin, "Admin User"),
        (librarian, "Librarian One"),
        (librarian2, "Librarian Two"),
        (patron1, "Patron One"),
        (patron2, "Patron Two"),
        (patron3, "Patron Three"),
    ]:
        profile = UserProfile(
            **dict(
                user_id=patron.id,
                _displayname="id_" + str(patron.id),
                full_name=name,
            )
        )
        db.session.add(profile)
    db.session.commit()

    return {
        "admin": admin,
        "librarian": librarian,
        "librarian2": librarian2,
        "patron1": patron1,
        "patron2": patron2,
        "patron3": patron3,
    }


@pytest.fixture(scope="module")
def app_with_notifs(app):
    """App with notifications test templates."""
    app.register_blueprint(
        Blueprint("invenio_app_ils_tests", __name__, template_folder="templates")
    )
    # add extra test templates to the search app blueprint, to fake the
    # existence of `invenio-theme` base templates.
    test_templates_path = os.path.join(os.path.dirname(__file__), "templates")
    enhanced_jinja_loader = jinja2.ChoiceLoader(
        [
            app.jinja_loader,
            jinja2.FileSystemLoader(test_templates_path),
        ]
    )
    # override default app jinja_loader to add the new path
    app.jinja_loader = enhanced_jinja_loader
    yield app


@pytest.fixture(scope="module")
def create_app():
    """Create test app."""
    return _create_app


@pytest.fixture(scope="module")
def base_app(create_app, app_config, request, default_handler):
    """Base application fixture (without database, search and cache).

    Scope: module.

    This fixture is responsible for creating the Invenio application. It
    depends on an application factory fixture that must be defined by the user.

    .. code-block:: python

        # confest.py
        import pytest

        @pytest.fixture(scope='module)
        def create_app():
            from invenio_app.factory import create_api
            return create_api

    It is possible to overide the application factory for a specific test
    module, either by defining a fixture like above example, or simply setting
    the ``create_app`` property on the module:

    .. code-block:: python

        # test_something.py

        from invenio_app.factory import create_api
        create_app = create_api

        def test_acase(base_app):
            # ...
    """
    # Use create_app from the module if defined, otherwise use default
    # create_app fixture.
    create_app = getattr(request.module, "create_app", create_app)
    app_ = create_app(**app_config)

    def delete_user_from_g(exception):
        """Delete user from `flask.g` when the request is tearing down.

        Flask-login==0.6.2 changed the way the user is saved i.e uses `flask.g`.
        Flask.g is pointing to the application context which is initialized per
        request. That said, `pytest-flask` is pushing an application context on each
        test initialization that causes problems as subsequent requests during a test
        are detecting the active application request and not popping it when the
        sub-request is tearing down. That causes the logged in user to remain cached
        for the whole duration of the test. To fix this, we add an explicit teardown
        handler that will pop out the logged in user in each request and it will force
        the user to be loaded each time.
        """
        from flask import g

        if "_login_user" in g:
            del g._login_user

    app_.teardown_request(delete_user_from_g)

    # See documentation for default_handler
    if default_handler:
        app_.logger.addHandler(default_handler)
    yield app_



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Loan indexer APIs."""

from datetime import datetime

from celery import shared_task
from flask import current_app
from invenio_circulation.pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from invenio_circulation.proxies import current_circulation
from invenio_indexer.api import RecordIndexer
from invenio_pidstore.errors import PIDDeletedError

from invenio_app_ils.circulation.utils import resolve_item_from_loan
from invenio_app_ils.documents.api import DOCUMENT_PID_TYPE
from invenio_app_ils.indexer import ReferencedRecordsIndexer
from invenio_app_ils.proxies import current_app_ils


@shared_task(ignore_result=True)
def index_referenced_records(loan):
    """Index referenced records."""
    indexer = ReferencedRecordsIndexer()
    indexed = dict(pid_type=CIRCULATION_LOAN_PID_TYPE, record=loan)
    referenced = []

    # fetch and index the document
    document_cls = current_app_ils.document_record_cls
    document = document_cls.get_record_by_pid(loan["document_pid"])
    referenced.append(dict(pid_type=DOCUMENT_PID_TYPE, record=document))

    # fetch and index the item
    if loan.get("item_pid"):
        item = resolve_item_from_loan(loan["item_pid"])
        referenced.append(dict(pid_type=loan["item_pid"]["type"], record=item))

    # index the loan itself: this is needed because of the extra field
    # `available_items_for_loan_count` added when indexing.
    # To calculate the value of this field, a search on the `loans`
    # indexed is performed and this loan has to be already indexed
    # with its latest data.
    # At the first indexing, `available_items_for_loan_count` value might
    # be wrong and corrected at the second re-indexing.
    loan_class = current_circulation.loan_record_cls
    loan_record = loan_class.get_record_by_pid(loan["pid"])

    referenced.append(dict(pid_type=CIRCULATION_LOAN_PID_TYPE, record=loan_record))

    # add all the other loans, as after indexing this one, they
    # will be affected in search
    pending_loans = document.search_loan_references().scan()

    for loan_hit in pending_loans:
        pending_loan = loan_class.get_record_by_pid(loan_hit.pid)
        referenced.append(dict(pid_type=CIRCULATION_LOAN_PID_TYPE, record=pending_loan))

    # index the loan and referenced records
    indexer.index(indexed, referenced)


class LoanIndexer(RecordIndexer):
    """Indexer class for Loan record."""

    def index(self, loan, arguments=None, **kwargs):
        """Index an Loan."""
        super().index(loan)
        eta = datetime.utcnow() + current_app.config["ILS_INDEXER_TASK_DELAY"]
        index_referenced_records.apply_async((loan,), eta=eta)


def index_extra_fields_for_loan(loan_dict):
    """Indexer hook to modify the loan record dict before indexing.

    The `available_items_for_loan_count` and `can_circulate_items_count` fields
    are added to the loan only on the search index because they are needed for
    search aggregation/filtering. They are not needed when fetching the loan
    details.
    """
    document_class = current_app_ils.document_record_cls
    try:
        document_record = document_class.get_record_by_pid(loan_dict["document_pid"])
    except PIDDeletedError:
        # Document might have been deleted while reindexing asynchronously.
        return

    document = document_record.replace_refs()

    items_available_for_loan_count = document["circulation"][
        "available_items_for_loan_count"
    ]
    loan_dict["available_items_for_loan_count"] = items_available_for_loan_count

    can_circulate_items_count = document["circulation"]["can_circulate_items_count"]
    loan_dict["can_circulate_items_count"] = can_circulate_items_count



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Series permissions."""

import json

from flask import url_for

from tests.helpers import user_login

_HTTP_OK = [200, 201, 204]
SERIES_PID = "serid-1"
ITEM_ENDPOINT = "invenio_records_rest.serid_item"
LIST_ENDPOINT = "invenio_records_rest.serid_list"


def test_series_read_permissions(client, testdata, json_headers, users):
    """Test Series endpoints read permissions."""
    tests = [
        ("admin", _HTTP_OK),
        ("librarian", _HTTP_OK),
        ("patron1", _HTTP_OK),
        ("anonymous", _HTTP_OK),
    ]

    def _test_list(expected_status):
        """Test get list."""
        url = url_for(LIST_ENDPOINT)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_read(expected_status, pid):
        """Test record read."""
        url = url_for(ITEM_ENDPOINT, pid_value=pid)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    for username, expected_status in tests:
        user_login(client, username, users)
        _test_list(expected_status)
        _test_read(expected_status, SERIES_PID)


def test_series_edit_permissions(client, testdata, json_headers, users):
    """Test series endpoints permissions."""
    dummy_series = dict(
        title="The Gulf: The Making of An American Sea",
        authors=["Einstein, Albert", "Stachel, John J et al."],
        mode_of_issuance="MULTIPART_MONOGRAPH",
    )
    tests = [
        ("admin", _HTTP_OK, dummy_series),
        ("librarian", _HTTP_OK, dummy_series),
        ("patron1", [403], dummy_series),
        ("anonymous", [401], dummy_series),
    ]

    def _test_create(expected_status, data):
        """Test record creation."""
        url = url_for(LIST_ENDPOINT)
        res = client.post(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status

        if res.status_code < 400:
            record = res.get_json()["metadata"]
            assert record
            return record["pid"]

    def _test_update(expected_status, data, pid):
        """Test record update."""
        pid_value = pid or SERIES_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.put(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status
        if res.status_code < 400:
            record = res.get_json()["metadata"]
            assert record

    def _test_delete(expected_status, pid):
        """Test record delete."""
        pid_value = pid or SERIES_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.delete(url, headers=json_headers)
        assert res.status_code in expected_status

    for username, expected_status, data in tests:
        user_login(client, username, users)
        pid = _test_create(expected_status, data)
        _test_update(expected_status, data, pid)
        _test_delete(expected_status, pid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Configuration for Invenio ILS acquisition module."""
from invenio_indexer.api import RecordIndexer
from invenio_records_rest.facets import terms_filter

from invenio_app_ils.config import RECORDS_REST_MAX_RESULT_WINDOW
from invenio_app_ils.permissions import backoffice_permission, superuser_permission

from .api import ORDER_PID_FETCHER, ORDER_PID_MINTER, ORDER_PID_TYPE, Order
from .search import OrderSearch

_ORDER_CONVERTER = 'pid(acqoid, record_class="invenio_app_ils.acquisition.api:Order")'

RECORDS_REST_ENDPOINTS = dict(
    acqoid=dict(
        pid_type=ORDER_PID_TYPE,
        pid_minter=ORDER_PID_MINTER,
        pid_fetcher=ORDER_PID_FETCHER,
        search_class=OrderSearch,
        record_class=Order,
        indexer_class=RecordIndexer,
        record_loaders={
            "application/json": "invenio_app_ils.acquisition.loaders:order_loader"
        },
        record_serializers={
            "application/json": "invenio_app_ils.records.serializers:json_v1_response"
        },
        search_serializers={
            "application/json": "invenio_app_ils.records.serializers:json_v1_search",
            "text/csv": "invenio_app_ils.records.serializers:csv_v1_search",
        },
        search_serializers_aliases={
            "csv": "text/csv",
            "json": "application/json",
        },
        list_route="/acquisition/orders/",
        item_route="/acquisition/orders/<{0}:pid_value>".format(_ORDER_CONVERTER),
        default_media_type="application/json",
        max_result_window=RECORDS_REST_MAX_RESULT_WINDOW,
        error_handlers=dict(),
        read_permission_factory_imp=backoffice_permission,
        list_permission_factory_imp=backoffice_permission,
        create_permission_factory_imp=backoffice_permission,
        update_permission_factory_imp=backoffice_permission,
        delete_permission_factory_imp=superuser_permission,
    ),
)

RECORDS_REST_SORT_OPTIONS = dict(
    acq_orders=dict(  # OrderSearch.Meta.index
        created=dict(fields=["_created"], title="Recently added", order=1),
        bestmatch=dict(
            fields=["-_score"],
            title="Best match",
            order=2,
        ),
        order_date=dict(
            fields=["order_date"],
            title="Order date",
            order=3,
        ),
        expected_delivery_date=dict(
            fields=["expected_delivery_date"],
            title="Expected delivery date",
            order=4,
        ),
        grand_total=dict(
            fields=["grand_total_main_currency.value"],
            title="Total",
            order=5,
        ),
    ),
)

FACET_PROVIDER_LIMIT = 5

RECORDS_REST_FACETS = dict(
    acq_orders=dict(  # OrderSearch.Meta.index
        aggs=dict(
            status=dict(terms=dict(field="status")),
            provider=dict(
                terms=dict(field="provider.name.keyword", size=FACET_PROVIDER_LIMIT)
            ),
            payment_mode=dict(terms=dict(field="order_lines.payment_mode")),
            medium=dict(terms=dict(field="order_lines.medium")),
        ),
        post_filters=dict(
            status=terms_filter("status"),
            provider=terms_filter("provider.name.keyword"),
            payment_mode=terms_filter("order_lines.payment_mode"),
            medium=terms_filter("order_lines.medium"),
        ),
    )
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Simple test of correct patron redirection after being logged out."""

from flask import url_for

from tests.helpers import user_login


def test_correct_redirection_after_logout(client, users):
    """
    Test that a patron is redirected into /logged-out page after logging out.
    """
    user_login(client, "patron1", users)
    res = client.get(url_for("security.logout"), follow_redirects=True)

    assert res.status_code == 200
    assert res.request.path == url_for("logged_out.logged_out_view")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""ILS Document APIs."""

from functools import partial
from urllib.parse import quote_plus

from flask import current_app
from invenio_circulation.search.api import search_by_pid
from invenio_pidstore.errors import PersistentIdentifierError
from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2

from invenio_app_ils.acquisition.proxies import current_ils_acq
from invenio_app_ils.errors import RecordHasReferencesError
from invenio_app_ils.fetchers import pid_fetcher
from invenio_app_ils.ill.proxies import current_ils_ill
from invenio_app_ils.minters import pid_minter
from invenio_app_ils.proxies import current_app_ils
from invenio_app_ils.records_relations.api import IlsRecordWithRelations
from invenio_app_ils.vocabularies.jsonresolvers.licenses import (
    REF_LICENSE_URL_ABSOLUTE_PATH,
)

DOCUMENT_PID_TYPE = "docid"
DOCUMENT_PID_MINTER = "docid"
DOCUMENT_PID_FETCHER = "docid"

DocumentIdProvider = type(
    "DocumentIdProvider",
    (RecordIdProviderV2,),
    dict(pid_type=DOCUMENT_PID_TYPE, default_status=PIDStatus.REGISTERED),
)
document_pid_minter = partial(pid_minter, provider_cls=DocumentIdProvider)
document_pid_fetcher = partial(pid_fetcher, provider_cls=DocumentIdProvider)


class Document(IlsRecordWithRelations):
    """Document record class."""

    DOCUMENT_TYPES = [
        "BOOK",
        "PROCEEDINGS",
        "STANDARD",
        "SERIAL_ISSUE",
        "ARTICLE",
        "MULTIMEDIA",
    ]

    _pid_type = DOCUMENT_PID_TYPE
    _schema = "documents/document-v2.0.0.json"
    _circulation_resolver_path = (
        "{scheme}://{host}/api/resolver/documents/{document_pid}/circulation"
    )
    _item_resolver_path = (
        "{scheme}://{host}/api/resolver/documents/{document_pid}/items"
    )
    _eitem_resolver_path = (
        "{scheme}://{host}/api/resolver/documents/{document_pid}/eitems"
    )
    _relations_path = (
        "{scheme}://{host}/api/resolver/documents/{document_pid}/relations"
    )
    _stock_resolver_path = (
        "{scheme}://{host}/api/resolver/documents/{document_pid}/stock"
    )
    _license_resolver_path = REF_LICENSE_URL_ABSOLUTE_PATH

    @classmethod
    def build_resolver_fields(cls, data):
        """Build all resolver fields."""
        data["circulation"] = {
            "$ref": cls._circulation_resolver_path.format(
                scheme=current_app.config["JSONSCHEMAS_URL_SCHEME"],
                host=current_app.config["JSONSCHEMAS_HOST"],
                document_pid=data["pid"],
            )
        }
        data.setdefault("relations", {})
        data["relations"] = {
            "$ref": cls._relations_path.format(
                scheme=current_app.config["JSONSCHEMAS_URL_SCHEME"],
                host=current_app.config["JSONSCHEMAS_HOST"],
                document_pid=data["pid"],
            )
        }
        data.setdefault("eitems", {})
        data["eitems"] = {
            "$ref": cls._eitem_resolver_path.format(
                scheme=current_app.config["JSONSCHEMAS_URL_SCHEME"],
                host=current_app.config["JSONSCHEMAS_HOST"],
                document_pid=data["pid"],
            )
        }
        data.setdefault("items", {})
        data["items"] = {
            "$ref": cls._item_resolver_path.format(
                scheme=current_app.config["JSONSCHEMAS_URL_SCHEME"],
                host=current_app.config["JSONSCHEMAS_HOST"],
                document_pid=data["pid"],
            )
        }
        data["stock"] = {
            "$ref": cls._stock_resolver_path.format(
                scheme=current_app.config["JSONSCHEMAS_URL_SCHEME"],
                host=current_app.config["JSONSCHEMAS_HOST"],
                document_pid=data["pid"],
            )
        }

        for lic in data.get("licenses", []):
            license_id = quote_plus(lic["license"]["id"])
            lic["license"]["$ref"] = cls._license_resolver_path.format(
                scheme=current_app.config["JSONSCHEMAS_URL_SCHEME"],
                host=current_app.config["JSONSCHEMAS_HOST"],
                license_id=license_id,
            )

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """Create Document record."""
        cls.build_resolver_fields(data)
        return super().create(data, id_=id_, **kwargs)

    def update(self, *args, **kwargs):
        """Update Document record."""
        super().update(*args, **kwargs)
        self.build_resolver_fields(self)

    def search_loan_references(self):
        """Search for referencing loans."""
        loan_search_res = search_by_pid(
            document_pid=self["pid"],
            filter_states=current_app.config["CIRCULATION_STATES_LOAN_REQUEST"]
            + current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"],
        )
        return loan_search_res

    def search_item_references(self):
        """Search for referencing items."""
        item_search = current_app_ils.item_search_cls()
        return item_search.search_by_document_pid(document_pid=self["pid"])

    def search_eitem_references(self):
        """Search for referencing eitems."""
        eitem_search = current_app_ils.eitem_search_cls()
        return eitem_search.search_by_document_pid(document_pid=self["pid"])

    def search_doc_req_references(self):
        """Search for referencing document requests."""
        req_search = current_app_ils.document_request_search_cls()
        return req_search.search_by_document_pid(document_pid=self["pid"])

    def search_order_references(self):
        """Search for referencing orders."""
        order_search = current_ils_acq.order_search_cls()
        return order_search.search_by_document_pid(
            document_pid=self["pid"],
        )

    def search_brw_req_references(self):
        """Search for referencing borrowing requests."""
        brw_req_search = current_ils_ill.borrowing_request_search_cls()
        return brw_req_search.search_by_document_pid(
            document_pid=self["pid"],
        )

    def has_references(self):
        """Check if record has references."""
        return any(
            [
                self.search_loan_references().count(),
                self.search_item_references().count(),
                self.search_eitem_references().count(),
                self.search_doc_req_references().count(),
                self.search_order_references().count(),
                self.search_brw_req_references().count(),
            ]
        )

    def delete(self, **kwargs):
        """Delete Document record."""
        loan_search_res = self.search_loan_references()
        item_search_res = self.search_item_references()
        eitem_search_res = self.search_eitem_references()
        req_search_res = self.search_doc_req_references()
        orders_refs_search = self.search_order_references()
        brw_req_refs_search = self.search_brw_req_references()

        if self.search_loan_references().count():
            raise RecordHasReferencesError(
                record_type="Document",
                record_id=self["pid"],
                ref_type="Loan",
                ref_ids=sorted([res["pid"] for res in loan_search_res.scan()]),
            )

        if item_search_res.count():
            raise RecordHasReferencesError(
                record_type="Document",
                record_id=self["pid"],
                ref_type="Item",
                ref_ids=sorted([res["pid"] for res in item_search_res.scan()]),
            )

        if eitem_search_res.count():
            raise RecordHasReferencesError(
                record_type="Document",
                record_id=self["pid"],
                ref_type="EItem",
                ref_ids=sorted([res["pid"] for res in eitem_search_res.scan()]),
            )

        if req_search_res.count():
            raise RecordHasReferencesError(
                record_type="Document",
                record_id=self["pid"],
                ref_type="DocumentRequest",
                ref_ids=sorted([res["pid"] for res in req_search_res.scan()]),
            )

        if orders_refs_search.count():
            raise RecordHasReferencesError(
                record_type="Document",
                record_id=self["pid"],
                ref_type="AcquisitionOrder",
                ref_ids=sorted([res["pid"] for res in orders_refs_search.scan()]),
            )

        if brw_req_refs_search.count():
            raise RecordHasReferencesError(
                record_type="Document",
                record_id=self["pid"],
                ref_type="BorrowingRequest",
                ref_ids=sorted([res["pid"] for res in brw_req_refs_search.scan()]),
            )
        return super().delete(**kwargs)


def document_exists(document_pid):
    """Return True if the Document exists given a PID."""
    Document = current_app_ils.document_record_cls
    try:
        Document.get_record_by_pid(document_pid)
    except PersistentIdentifierError:
        return False
    return True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2024 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""invenio-app-ils."""

__version__ = "4.4.0"

__all__ = ("__version__",)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Ils Records API."""

from flask import current_app
from invenio_jsonschemas import current_jsonschemas
from invenio_pidstore.models import PersistentIdentifier
from invenio_pidstore.resolver import Resolver
from invenio_records.api import Record
from invenio_rest.errors import FieldError
from jsonschema.exceptions import ValidationError
from werkzeug.utils import cached_property

from invenio_app_ils.errors import IlsValidationError


class RecordValidator(object):
    """ILS record validator."""

    def validate(self, record, **kwargs):
        """Validate record before create and commit.

        :param record: Record to be validated.
        :param kwargs: Keyword arguments passed from RecordBase:validate.
        """
        pass


class IlsRecord(Record):
    """Ils record class."""

    CURATOR_TYPES = ["user_id", "script", "import"]

    _validator = RecordValidator()

    @cached_property
    def pid(self):
        """Get the PersistentIdentifier for this record."""
        return PersistentIdentifier.get(pid_type=self._pid_type, pid_value=self["pid"])

    @staticmethod
    def pid_type_to_record_class(pid_type):
        """Get the record class from the pid_type."""
        endpoints_cfg = current_app.config["RECORDS_REST_ENDPOINTS"]
        return endpoints_cfg[pid_type]["record_class"]

    @classmethod
    def get_record_by_pid(cls, pid_value, with_deleted=False, pid_type=None):
        """Get ils record by pid value."""
        if pid_type is None:
            pid_type = cls._pid_type
        else:
            new_cls = cls.pid_type_to_record_class(pid_type)
            return new_cls.get_record_by_pid(pid_value, with_deleted=with_deleted)

        resolver = Resolver(pid_type=pid_type, object_type="rec", getter=cls.get_record)
        _, record = resolver.resolve(str(pid_value))
        return record

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """Create IlsRecord record."""
        data["$schema"] = current_jsonschemas.path_to_url(cls._schema)
        return super().create(data, id_=id_, **kwargs)

    def clear(self):
        """Clear IlsRecord data."""
        super().clear()
        self["$schema"] = current_jsonschemas.path_to_url(self._schema)

    def _validate(self, **kwargs):
        """Validate ILS record."""
        # JSON schema validation
        try:
            json = super()._validate(**kwargs)
        except ValidationError as jve:
            path = ".".join(str(x) for x in jve.path)
            errors = [FieldError(path, jve.message)]
            raise IlsValidationError(
                description="Record validation error",
                errors=errors,
                original_exception=jve,
            )

        # Custom record validation
        if self._validator:
            self._validator.validate(self, **kwargs)
        return json



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Loan custom serializer functions."""

from invenio_circulation.proxies import current_circulation
from invenio_pidstore.errors import PIDDeletedError, PIDDoesNotExistError

from invenio_app_ils.records.jsonresolvers.api import pick


def field_loan(metadata):
    """Get the loan object and add it to the metadata."""
    loan_pid = metadata.get("patron_loan", {}).get("pid")
    if not loan_pid:
        return
    Loan = current_circulation.loan_record_cls
    try:
        loan = Loan.get_record_by_pid(loan_pid)
    except PIDDeletedError:
        metadata["patron_loan"]["loan"] = {"pid": "This loan was deleted."}
        return
    except PIDDoesNotExistError:
        metadata["patron_loan"]["loan"] = {"pid": "Invalid Loan PID."}
        return
    metadata["patron_loan"]["loan"] = pick(
        loan, "pid", "start_date", "end_date", "state", "extension_count"
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolve the relations referenced in the Series."""

import jsonresolver
from werkzeug.routing import Rule

from invenio_app_ils.records.jsonresolvers.api import get_pid_or_default
from invenio_app_ils.series.api import Series

# Note: there must be only one resolver per file,
# otherwise only the last one is registered


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the relations of a Series record."""
    from flask import current_app

    @get_pid_or_default(default_value=dict())
    def relations_resolver(series_pid):
        """Resolve record relations."""
        series = Series.get_record_by_pid(series_pid)
        return series.relations

    url_map.add(
        Rule(
            "/api/resolver/series/<series_pid>/relations",
            endpoint=relations_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS Patrons views."""

from flask import Blueprint, current_app
from flask_login import current_user
from invenio_circulation.search.api import search_by_patron_item_or_document
from webargs import fields
from webargs.flaskparser import use_kwargs

from invenio_app_ils.circulation.search import get_loans_aggregated_by_states
from invenio_app_ils.permissions import need_permissions


def get_user_loan_information_blueprint(_):
    """Add patron views to the blueprint."""
    blueprint = Blueprint("invenio_app_ils_patrons", __name__)

    @blueprint.route("/me/loans")
    @use_kwargs({"document_pid": fields.Str(required=True)})
    @need_permissions("patron-loans")
    def get_user_information(document_pid):
        patron_pid = str(current_user.id)
        return retrieve_user_loans_information(patron_pid, document_pid)

    return blueprint


def retrieve_user_loans_information(patron_pid, document_pid):
    """Retrieves patron loans for the given patron."""
    active_requested_loan_states = (
        current_app.config["CIRCULATION_STATES_LOAN_REQUEST"]
        + current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
    )
    past_loan_states = current_app.config["CIRCULATION_STATES_LOAN_COMPLETED"]

    user_information = {
        "is_requested": False,
        "has_active_loan": False,
        "last_loan": None,
    }

    loans_search = get_loans_aggregated_by_states(
        document_pid,
        active_requested_loan_states + past_loan_states,
        patron_pid,
    )
    # No need for the loan hits
    loans_search = loans_search[:0]
    loan_result = loans_search.execute()
    for bucket in loan_result.aggregations.states.buckets:
        if (
            bucket["key"] in current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
            and bucket["doc_count"] > 0
        ):
            user_information["has_active_loan"] = True
        elif (
            bucket["key"] in current_app.config["CIRCULATION_STATES_LOAN_REQUEST"]
            and bucket["doc_count"] > 0
        ):
            user_information["is_requested"] = True
        elif (
            bucket["key"] in current_app.config["CIRCULATION_STATES_LOAN_COMPLETED"]
            and bucket["doc_count"] > 0
        ):
            search = search_by_patron_item_or_document(
                patron_pid=patron_pid,
                document_pid=document_pid,
                filter_states=past_loan_states,
            )

            search = search.sort({"end_date": {"order": "desc"}})
            search_result = search.execute()
            has_past_loans = search_result.hits.total.value > 0
            if has_past_loans:
                last_loan = search_result.hits[0]
                user_information["last_loan"] = last_loan["end_date"]

    return user_information



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Ils Locations API."""

from functools import partial

from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2

from invenio_app_ils.errors import RecordHasReferencesError
from invenio_app_ils.fetchers import pid_fetcher
from invenio_app_ils.minters import pid_minter
from invenio_app_ils.proxies import current_app_ils
from invenio_app_ils.records.api import IlsRecord

LOCATION_PID_TYPE = "locid"
LOCATION_PID_MINTER = "locid"
LOCATION_PID_FETCHER = "locid"

LocationIdProvider = type(
    "LocationIdProvider",
    (RecordIdProviderV2,),
    dict(pid_type=LOCATION_PID_TYPE, default_status=PIDStatus.REGISTERED),
)
location_pid_minter = partial(pid_minter, provider_cls=LocationIdProvider)
location_pid_fetcher = partial(pid_fetcher, provider_cls=LocationIdProvider)


class Location(IlsRecord):
    """Location record class."""

    _pid_type = LOCATION_PID_TYPE
    _schema = "locations/location-v1.0.0.json"

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """Create Location record."""
        return super().create(data, id_=id_, **kwargs)

    def delete(self, **kwargs):
        """Delete Location record."""
        iloc_search = current_app_ils.internal_location_search_cls()
        iloc_search_res = iloc_search.search_by_location_pid(location_pid=self["pid"])
        if iloc_search_res.count():
            raise RecordHasReferencesError(
                record_type="Location",
                record_id=self["pid"],
                ref_type="Internal Location",
                ref_ids=sorted([res["pid"] for res in iloc_search_res.scan()]),
            )
        rec = super().delete(**kwargs)

        try:
            # invalidate cached property
            del current_app_ils.get_default_location_pid
        except (AttributeError, KeyError):
            pass

        return rec



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolve the Loan referenced in the Item."""

import jsonresolver
from invenio_circulation.api import get_loan_for_item
from werkzeug.routing import Rule

from invenio_app_ils.items.api import ITEM_PID_TYPE

# Note: there must be only one resolver per file,
# otherwise only the last one is registered


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the referred Loan for an Item record."""
    from flask import current_app

    def loan_for_item_resolver(item_pid):
        """Return the loan for the given item."""
        loan = get_loan_for_item(dict(value=item_pid, type=ITEM_PID_TYPE))
        if not loan:
            return {}
        else:
            return {
                "loan_pid": loan["pid"],
                "patron_pid": loan["patron_pid"],
                "patron": loan["patron"],
                "document_pid": loan["document_pid"],
                "item_pid": loan.get("item_pid"),
                "state": loan["state"],
                "start_date": loan["start_date"],
                "end_date": loan["end_date"],
                "extension_count": loan.get("extension_count", 0),
                "is_overdue": loan.get("is_overdue", False),
            }

    url_map.add(
        Rule(
            "/api/resolver/items/<item_pid>/loan",
            endpoint=loan_for_item_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolve the EItem referenced in the Document."""

import jsonresolver
from werkzeug.routing import Rule

from invenio_app_ils.eitems.api import EItem
from invenio_app_ils.proxies import current_app_ils

# Note: there must be only one resolver per file,
# otherwise only the last one is registered


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the referred EItems for a Document record."""
    from flask import current_app

    def eitems_resolver(document_pid):
        """Search and return the EItems that reference this Document."""
        eitems = []
        eitem_search = current_app_ils.eitem_search_cls()
        for hit in eitem_search.search_by_document_pid(document_pid).scan():
            eitem = hit.to_dict()
            eitems.append(
                {
                    "pid": eitem.get("pid"),
                    "eitem_type": eitem.get("eitem_type"),
                    "description": eitem.get("description"),
                    "identifiers": eitem.get("identifiers", []),
                    "internal_notes": eitem.get("internal_notes"),
                    "open_access": eitem.get("open_access"),
                    "bucket_id": eitem.get("bucket_id", None),
                    "files": eitem.get("files", []),
                    "urls": eitem.get("urls", []),
                }
            )

        return {"total": len(eitems), "hits": eitems}

    url_map.add(
        Rule(
            "/api/resolver/documents/<document_pid>/eitems",
            endpoint=eitems_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Document requests notifications APIs."""

from flask import current_app
from invenio_records_rest.utils import obj_or_import_string

from invenio_app_ils.ill.errors import ILLError
from invenio_app_ils.notifications.api import build_common_msg_ctx, send_notification


def _build_ill_msg_ctx(brw_req):
    """Return a dict with extra ILL ctx to pass to jinja templates."""
    try:
        # fetch the patron loan if it exists
        loan = brw_req.patron_loan.get()
    except ILLError:
        # no loan in the borrowing request
        loan = dict()

    return dict(patron_loan=loan)


def send_ill_notification(brw_req, action, msg_extra_ctx=None, **kwargs):
    """Send notification to patron about a document request action."""
    _filter = current_app.config["ILS_ILL_NOTIFICATIONS_FILTER"]
    if _filter and _filter(brw_req, action, **kwargs) is False:
        return

    msg_ctx = msg_extra_ctx or {}
    msg_ctx.update(build_common_msg_ctx(brw_req))
    msg_ctx.update(_build_ill_msg_ctx(brw_req))

    func_or_path = current_app.config["ILS_ILL_NOTIFICATIONS_MSG_BUILDER"]
    builder = obj_or_import_string(func_or_path)
    msg = builder(record=brw_req, action=action, msg_ctx=msg_ctx, **kwargs)

    patron = msg_ctx["patron"]
    patrons = [patron]

    send_notification(patrons, msg, **kwargs)


def ill_notifications_filter(record, action, **kwargs):
    """Filter ILL notifications to be sent."""
    return True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test ILL borrowing requests actions."""

import json
from datetime import timedelta

import arrow
from flask import url_for

from invenio_app_ils.ill.api import BORROWING_REQUEST_PID_TYPE
from tests.helpers import user_login, user_logout


def _create_on_loan_brwreq(
    patron_id, loan_start_date, loan_end_date, client, json_headers
):
    """Create a new ON_LOAN ILL borrowing request."""

    def _create_brwreq(patron_id):
        brwreq = dict(
            document_pid="docid-3",
            provider_pid="ill-provid-2",
            patron_pid=patron_id,
            status="REQUESTED",
            type="PHYSICAL_COPY",
        )
        url = url_for("invenio_records_rest.illbid_list")
        res = client.post(url, headers=json_headers, data=json.dumps(brwreq))
        assert res.status_code == 201
        brw_req = res.get_json()["metadata"]
        return brw_req["pid"]

    def _create_loan_action(pid):
        """Send request to create a new loan from the ILL."""
        data = dict(
            loan_start_date=loan_start_date,
            loan_end_date=loan_end_date,
            transaction_location_pid="locid-1",
        )
        url = url_for("invenio_app_ils_ill.illbid_create_loan", pid_value=pid)
        res = client.post(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code == 200
        brw_req = res.get_json()["metadata"]
        assert brw_req["status"] == "ON_LOAN"
        return brw_req

    brwreq_pid = _create_brwreq(patron_id)
    return _create_loan_action(brwreq_pid)


def _create_on_loan_brwreq_random_dates(patron_id, client, json_headers):
    """Create a new ON_LOAN ILL borrowing request with random start/end."""
    loan_start_date = arrow.utcnow()
    loan_end_date = loan_start_date + timedelta(days=15)
    brwreq = _create_on_loan_brwreq(
        patron_id,
        loan_start_date.date().isoformat(),
        loan_end_date.date().isoformat(),
        client,
        json_headers,
    )
    return brwreq, brwreq["pid"]


def _create_on_loan_brwreq_with_pending_extension(patron_id, client, json_headers):
    """Create a new ON_LOAN ILL borrowing request with pending extension."""
    today = arrow.utcnow().date().isoformat()
    brwreq, brwreq_pid = _create_on_loan_brwreq_random_dates("1", client, json_headers)
    res = _request_extension_action(brwreq_pid, client, json_headers)
    assert res.status_code == 200

    brwreq = res.get_json()["metadata"]
    patron_loan = brwreq["patron_loan"]

    assert patron_loan["loan"]["state"] == "ITEM_ON_LOAN"
    assert "extension_count" not in patron_loan["loan"]
    assert patron_loan["extension"]["status"] == "PENDING"
    assert patron_loan["extension"]["request_date"] == today

    return brwreq, brwreq["pid"]


def _request_extension_action(pid, client, json_headers):
    """Send HTTP request to request extension of ILL Borrowing Request."""
    url = url_for("invenio_app_ils_ill.illbid_request_extension", pid_value=pid)
    return client.post(url, headers=json_headers, data=json.dumps({}))


def _accept_extension_action(pid, data, client, json_headers):
    """Send HTTP request to accept extension of ILL Borrowing Request."""
    data["transaction_location_pid"] = "locid-1"
    url = url_for("invenio_app_ils_ill.illbid_accept_extension", pid_value=pid)
    return client.post(url, headers=json_headers, data=json.dumps(data))


def _decline_extension_action(pid, _, client, json_headers):
    """Send HTTP request to decline extension of ILL Borrowing Request."""
    url = url_for("invenio_app_ils_ill.illbid_decline_extension", pid_value=pid)
    return client.post(url, headers=json_headers, data=json.dumps({}))


def test_brwreq_request_extension_only_owner(client, testdata, json_headers, users):
    """Test that patron can request extension only for his own loan."""
    user_login(client, "librarian", users)

    brwreq, brwreq_pid = _create_on_loan_brwreq_random_dates("1", client, json_headers)

    # anonymous, forbidden
    user_logout(client)
    res = _request_extension_action(brwreq_pid, client, json_headers)
    assert res.status_code == 401

    # patron2 is not the owner, forbidden
    user_login(client, "patron2", users)
    res = _request_extension_action(brwreq_pid, client, json_headers)
    assert res.status_code == 403

    # patron1 is the owner, success
    user_login(client, "patron1", users)
    res = _request_extension_action(brwreq_pid, client, json_headers)
    assert res.status_code == 200
    extension = res.get_json()["metadata"]["patron_loan"]["extension"]
    assert extension["status"] == "PENDING"
    assert extension["request_date"] == arrow.now().date().isoformat()

    # create a new one with patron2 owner, librarian can request extension
    user_login(client, "librarian", users)
    brwreq, brwreq_pid = _create_on_loan_brwreq_random_dates("2", client, json_headers)
    res = _request_extension_action(brwreq_pid, client, json_headers)
    assert res.status_code == 200
    assert extension["status"] == "PENDING"
    assert extension["request_date"] == arrow.now().date().isoformat()


def test_brwreq_accept_decline_extension_only_librarian(
    client, testdata, json_headers, users
):
    """Test that only librarian can accept or decline an extension."""

    tests = [_accept_extension_action, _decline_extension_action]
    for action in tests:
        user_login(client, "librarian", users)

        # create request
        brwreq, brwreq_pid = _create_on_loan_brwreq_with_pending_extension(
            "1", client, json_headers
        )

        loan_end_date = arrow.utcnow() + timedelta(days=15)
        data = dict(loan_end_date=loan_end_date.date().isoformat())

        # accept/decline

        # anonymous, forbidden
        user_logout(client)
        res = action(brwreq_pid, data, client, json_headers)
        assert res.status_code == 401

        # patron, forbidden
        user_login(client, "patron1", users)
        res = action(brwreq_pid, data, client, json_headers)
        assert res.status_code == 403

        # librarian, success
        user_login(client, "librarian", users)
        res = action(brwreq_pid, data, client, json_headers)
        assert res.status_code == 200


def test_brwreq_request_extension_fails_on_wrong_status(
    client, testdata, json_headers, users
):
    user_login(client, "patron1", users)
    # demo data "illbid-1" is not ON_LOAN
    res = _request_extension_action("illbid-1", client, json_headers)
    assert res.status_code == 400
    # demo data "illbid-7" has already a pending request extension
    res = _request_extension_action("illbid-7", client, json_headers)
    assert res.status_code == 400
    # demo data "illbid-" has declined request extension
    res = _request_extension_action("illbid-8", client, json_headers)
    assert res.status_code == 400


def test_brwreq_accept_decline_extension_should_fail_when_loan_not_active(
    client, testdata, json_headers, users
):
    """Test that accept or decline an extension fails on loan not active."""

    tests = [_accept_extension_action, _decline_extension_action]
    for action in tests:
        user_login(client, "librarian", users)

        # create request
        brwreq, brwreq_pid = _create_on_loan_brwreq_with_pending_extension(
            "1", client, json_headers
        )

        # check-in loan
        loan_pid = brwreq["patron_loan"]["pid"]
        url = url_for(
            "invenio_circulation_loan_actions.loanid_actions",
            pid_value=loan_pid,
            action="checkin",
        )
        item_pid = dict(type=BORROWING_REQUEST_PID_TYPE, value=brwreq["pid"])
        params = dict(
            document_pid=brwreq["document_pid"],
            item_pid=item_pid,
            patron_pid=brwreq["patron_pid"],
            transaction_location_pid="locid-1",
            transaction_user_pid="1",
        )
        res = client.post(url, headers=json_headers, data=json.dumps(params))
        assert res.status_code == 202

        # accept extension
        loan_end_date = arrow.utcnow() + timedelta(days=15)
        data = dict(loan_end_date=loan_end_date.date().isoformat())

        res = action(brwreq_pid, data, client, json_headers)
        assert res.status_code == 400


def test_brwreq_accept_extension_success(client, testdata, json_headers, users):
    """Test accept extension success."""
    user_login(client, "librarian", users)

    # create request
    brwreq, brwreq_pid = _create_on_loan_brwreq_with_pending_extension(
        "1", client, json_headers
    )

    loan_end_date = arrow.utcnow() + timedelta(days=15)
    end_date = loan_end_date.date().isoformat()
    data = dict(loan_end_date=end_date)

    user_login(client, "librarian", users)
    # accept extension
    res = _accept_extension_action(brwreq_pid, data, client, json_headers)
    assert res.status_code == 200

    patron_loan = res.get_json()["metadata"]["patron_loan"]
    assert patron_loan["loan"]["state"] == "ITEM_ON_LOAN"
    assert patron_loan["loan"]["extension_count"] == 1
    assert patron_loan["loan"]["end_date"] == end_date
    assert "status" not in patron_loan["extension"]


def test_brwreq_decline_extension_success(client, testdata, json_headers, users):
    """Test declone extension success."""
    user_login(client, "librarian", users)

    # create request
    brwreq, brwreq_pid = _create_on_loan_brwreq_with_pending_extension(
        "1", client, json_headers
    )
    end_date = brwreq["patron_loan"]["loan"]["end_date"]

    user_login(client, "librarian", users)
    # decline extension
    res = _decline_extension_action(brwreq_pid, dict(), client, json_headers)
    assert res.status_code == 200

    patron_loan = res.get_json()["metadata"]["patron_loan"]
    assert patron_loan["loan"]["state"] == "ITEM_ON_LOAN"
    assert "extension_count" not in patron_loan["loan"]
    assert patron_loan["loan"]["end_date"] == end_date
    assert patron_loan["extension"]["status"] == "DECLINED"

    # request again should now fail because it has been declined
    user_login(client, "patron1", users)
    res = _request_extension_action(brwreq_pid, client, json_headers)
    assert res.status_code == 400



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Items APIs."""

from invenio_app_ils.items.api import (
    ITEM_PID_TYPE,
    get_document_pid_by_item_pid,
    get_item_pids_by_document_pid,
    item_exists,
)


def test_get_item_pids_by_document_pid(testdata):
    """Test retrieve Items PIDs for the given Document PID."""
    first_doc_pid = testdata["documents"][0]["pid"]
    items_pids = get_item_pids_by_document_pid(first_doc_pid)
    assert len(list(items_pids)) == 9


def test_get_document_pid_by_item_pid(testdata):
    """Test retrieve Document PID for the given Item."""
    first_item_pid = testdata["items"][0]["pid"]
    item_pid = dict(type=ITEM_PID_TYPE, value=first_item_pid)
    doc_pid = get_document_pid_by_item_pid(item_pid)
    assert doc_pid == "docid-1"


def test_item_exists(testdata):
    """Test return True if item exists."""
    first_item_pid = testdata["items"][0]["pid"]
    item_pid = dict(type=ITEM_PID_TYPE, value=first_item_pid)
    assert item_exists(item_pid)


def test_item_not_exist(testdata):
    """Test return False if item does not exist."""
    item_pid = dict(type=ITEM_PID_TYPE, value="not-existing-item-pid")
    assert not item_exists(item_pid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolver for Document referenced in BorrowingRequest."""

import jsonresolver
from werkzeug.routing import Rule

from invenio_app_ils.documents.utils import flatten_authors
from invenio_app_ils.proxies import current_app_ils
from invenio_app_ils.records.jsonresolvers.api import (
    get_field_value_for_record as get_field_value,
)
from invenio_app_ils.records.jsonresolvers.api import get_pid_or_default, pick

from ..proxies import current_ils_ill

# Note: there must be only one resolver per file,
# otherwise only the last one is registered


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the referred Document for a Borrowing Request record."""
    from flask import current_app

    @get_pid_or_default(default_value=dict())
    def get_document(document_pid):
        """Return the Document record."""
        Document = current_app_ils.document_record_cls
        document = Document.get_record_by_pid(document_pid)
        obj = pick(
            document,
            "authors",
            "cover_metadata",
            "edition",
            "pid",
            "publication_year",
            "title",
            "alternative_titles",
        )
        obj["authors"] = flatten_authors(obj["authors"])
        return obj

    def borrowing_request_resolver(request_pid):
        """Return the Document record for the given Brw Request or raise."""
        request_record_cls = current_ils_ill.borrowing_request_record_cls
        document_pid = get_field_value(request_record_cls, request_pid, "document_pid")
        return get_document(document_pid)

    url_map.add(
        Rule(
            "/api/resolver/ill/borrowing-requests/<request_pid>/document",
            endpoint=borrowing_request_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test acquisition orders."""

import json

from flask import url_for

from tests.helpers import user_login

_HTTP_OK = [200, 201, 204]
ORDER_PID = "acqoid-1"
ITEM_ENDPOINT = "invenio_records_rest.acqoid_item"
LIST_ENDPOINT = "invenio_records_rest.acqoid_list"


def test_acq_orders_permissions(client, testdata, json_headers, users):
    """Test orders endpoints permissions."""
    dummy_acquisition_order = dict(
        status="PENDING",
        order_date="2020-02-25",
        provider_pid="acq-provid-1",
        order_lines=[
            dict(
                copies_ordered=3,
                document_pid="docid-1",
                medium="paper",
                recipient="library",
            )
        ],
    )

    def _test_list(expected_status):
        """Test get list."""
        url = url_for(LIST_ENDPOINT)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_create(expected_status, data, user):
        """Test record creation."""
        url = url_for(LIST_ENDPOINT)
        res = client.post(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status

        if res.status_code < 400:
            ord = res.get_json()["metadata"]
            assert ord["status"] == "PENDING"
            expected_created_by = dict(type="user_id", value=str(user.id))
            assert ord["created_by"] == expected_created_by
            assert not ord.get("updated_by")
            return ord["pid"]

    def _test_update(expected_status, data, pid, user):
        """Test record update."""
        pid_value = pid or ORDER_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.put(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status
        if res.status_code < 400:
            expected_changed_by = dict(type="user_id", value=str(user.id))
            ord = res.get_json()["metadata"]
            assert ord["created_by"] == expected_changed_by
            assert ord["updated_by"] == expected_changed_by

    def _test_read(expected_status, pid):
        """Test record read."""
        pid_value = pid or ORDER_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_delete(expected_status, pid):
        """Test record delete."""
        url = url_for(ITEM_ENDPOINT, pid_value=pid)
        res = client.delete(url, headers=json_headers)
        assert res.status_code in expected_status

    tests = [
        ("admin", _HTTP_OK, dummy_acquisition_order),
        ("librarian", _HTTP_OK, dummy_acquisition_order),
        ("patron1", [403], dummy_acquisition_order),
        ("anonymous", [401], dummy_acquisition_order),
    ]
    for username, expected_status, data in tests:
        user = user_login(client, username, users)
        _test_list(expected_status)
        pid = _test_create(expected_status, data, user)
        _test_update(expected_status, data, pid, user)
        _test_read(expected_status, pid)

    tests = [
        ("patron1", [403]),
        ("anonymous", [401]),
        ("librarian", [403]),
        ("admin", [204]),
    ]
    for username, expected_status in tests:
        user_login(client, username, users)
        _test_delete(expected_status, ORDER_PID)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test notifications."""

import pytest
from flask import url_for
from jinja2.exceptions import TemplateError, TemplateNotFound

from invenio_app_ils.notifications.api import (
    log_error_notification,
    log_successful_notification,
    send_notification,
)
from invenio_app_ils.notifications.messages import NotificationMsg
from invenio_app_ils.notifications.models import NotificationsLogs
from invenio_app_ils.patrons.api import Patron
from tests.helpers import user_login


def test_message_full(app_with_notifs):
    """Test that the title, body and html are read correctly."""
    with app_with_notifs.app_context():
        # remove footer
        app_with_notifs.config["ILS_NOTIFICATIONS_TEMPLATES"] = {}

        full = NotificationMsg("notifications/title_body_html.html")
        assert full.title == "Test title."
        assert "Test body." == full.body_plain
        assert "Test html." == full.body_html


def test_message_body_same_as_html(app_with_notifs):
    """Test that the title and body are read correctly."""
    with app_with_notifs.app_context():
        # remove footer
        app_with_notifs.config["ILS_NOTIFICATIONS_TEMPLATES"] = {}

        blank = NotificationMsg("notifications/title_body.html")
        assert blank.title == "Test title."
        assert "Test body." == blank.body_plain
        assert "Test body." == blank.body_html


def test_invalid_message_templates(app_with_notifs):
    """Test invalid templates."""
    with app_with_notifs.app_context():
        with pytest.raises(TemplateError) as ex:
            NotificationMsg("notifications/blank.html")
        assert "No block with name 'title'" in str(ex.value)

        with pytest.raises(TemplateError) as ex:
            NotificationMsg("notifications/title_only.html")
        assert "No block with name 'body_plain'" in str(ex.value)


def test_message_templates_escaping(app_with_notifs):
    """Test templates escaping."""
    title = (
        "Spèciâl chäráctèrs: "
        "`,~,!,@,#,$,%,^,&,*,(,),_,-,+,=,{,[,},},|,\\,:,;,',<,>,.,?,/,º,ª "
        "<script>alert('test');</script>"
    )
    ctx = dict(title=title)
    with app_with_notifs.app_context():
        # remove footer
        app_with_notifs.config["ILS_NOTIFICATIONS_TEMPLATES"] = {}

        full = NotificationMsg("notifications/title_body_html_ctx.html", ctx)

        assert full.title == 'Test "{0}" title.'.format(title)

        from markupsafe import escape

        escaped = escape(title)
        assert 'Test "{0}" body.'.format(escaped) == full.body_plain
        assert 'Test "{0}" html.'.format(escaped) == full.body_html


def test_message_with_missing_template(app_with_notifs):
    """Test that a missing template raises an exception."""
    with app_with_notifs.app_context():
        with pytest.raises(TemplateNotFound):
            NotificationMsg("missing.html")


def test_send_message_backend(app_with_notifs, users, testdata, mocker):
    """Test that notifications backend is called when sending."""
    send_mocked = mocker.patch("invenio_app_ils.notifications.backends.mail.send")
    send_mocked.__name__ = "send"
    send_mocked.__annotations__ = "send"
    backends = mocker.patch(
        "invenio_app_ils.notifications.api._get_notification_backends",
        return_value=[send_mocked],
    )
    with app_with_notifs.app_context():
        # remove footer
        app_with_notifs.config["ILS_NOTIFICATIONS_TEMPLATES"] = {}

        msg = NotificationMsg("notifications/title_body_html.html")
        patrons = [Patron(users["patron1"].id)]
        send_notification(patrons, msg)

        assert send_mocked.apply_async.called


class FakeMessage(NotificationMsg):
    def __init__(self, *args, **kwargs):
        template = kwargs.pop("template", "notifications/title_body_html.html")
        super().__init__(template, **kwargs)


def test_log_successful_error_mail_task(app_with_notifs, users, testdata, mocker):
    """Test that a successfully sent email is logged."""
    # !attention the mocker has to patch function where it is used, not defined
    succ = mocker.patch("invenio_app_ils.notifications.api.log_successful_notification")
    err = mocker.patch("invenio_app_ils.notifications.api.log_error_notification")

    patron = Patron(users["patron1"].id)

    assert not succ.s.called
    assert not err.s.called
    send_notification([patron], FakeMessage())
    assert succ.s.called
    assert err.s.called


def test_notifications_db_table_and_endpoint(
    users, client, json_headers, app_with_notifs
):
    """Test logging of notifs in db table and read from endpoint."""
    request = {"id": "test-id", "task": "test-task"}
    data = {
        "id": "test-id",
        "recipients": ["patron1@test.com"],
        "is_manually_triggered": True,
        "message_id": "1",
        "patron_id": "1",
    }
    exc = "An error occurred."
    log_successful_notification(None, data)

    log_error_notification(request=request, exc=exc, traceback=None, data=data)

    assert NotificationsLogs.query.filter_by(id=1).one().send_log == "Success"
    assert (
        NotificationsLogs.query.filter_by(id=2).one().send_log
        == "Error: 'An error occurred.'"
    )

    ITEM_ENDPOINT = "invenio_app_ils_notifications.get_notification"
    LIST_ENDPOINT = "invenio_app_ils_notifications.get_notifications"

    user_login(client, "librarian", users)
    url = url_for(LIST_ENDPOINT)
    res = client.get(url, headers=json_headers)
    assert res.get_json()["hits"][0]["send_log"] == "Error: 'An error occurred.'"
    assert res.get_json()["hits"][1]["send_log"] == "Success"

    url = url_for(ITEM_ENDPOINT, id=1)
    res = client.get(url, headers=json_headers)
    assert res.get_json()["send_log"] == "Success"
    assert res.get_json()["id"] == 1

    url = url_for(ITEM_ENDPOINT, id=2)
    res = client.get(url, headers=json_headers)

    assert res.get_json()["send_log"] == "Error: 'An error occurred.'"
    assert res.get_json()["id"] == 2
